
import datetime as dt
from logging import getLogger

from geoalchemy2 import Geography, Geometry
from sqlalchemy import Column, Text, Integer, ForeignKey, UniqueConstraint, desc, DateTime, Index, text
from sqlalchemy.orm import relationship, backref

from .source import SourceType, GroupedSource, Source
from ..support import Base
from ..triggers import add_child_ddl, add_text
from ..types import Sort, ShortCls, NullText, Name, name_and_title, Point, UTC
from ..utils import WGS84_SRID
from ...common.date import format_time, local_date_to_time, local_time_to_time
from ...lib.utils import timing

log = getLogger(__name__)


class ActivityGroup(Base):

    __tablename__ = 'activity_group'

    id = Column(Integer, primary_key=True)
    name = Column(Name, nullable=False, index=True, unique=True)
    title = Column(Text, nullable=False)
    description = Column(Text, nullable=False, server_default='')
    sort = Column(Sort, nullable=False)

    def __init__(self, **kargs):
        super().__init__(**name_and_title(kargs))

    def __str__(self):
        return self.name

    @classmethod
    def from_name(cls, s, name, none=False):
        if not name:
            return None
        elif isinstance(name, ActivityGroup):
            return name  # allow callers to already have an instance
        else:
            instance = s.query(ActivityGroup).filter(ActivityGroup.name.ilike(name)).one_or_none()
            if instance is None and not none:
                raise Exception(f'No activity group defined for {name}')
            return instance


@add_child_ddl(Source)
@add_text('''
alter table %(table)s
  add constraint no_activity_overlap
  exclude using gist (tstzrange(start, finish) with &&);
''')
class ActivityJournal(GroupedSource):

    __tablename__ = 'activity_journal'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    file_hash_id = Column(Integer, ForeignKey('file_hash.id'), nullable=False, index=True, unique=True)
    file_hash = relationship('FileHash', backref=backref('activity_journal', uselist=False))
    start = Column(UTC, nullable=False, index=True, unique=True)
    finish = Column(UTC, nullable=False)
    # nullable because created later
    centre = Column(Geography('Point', srid=WGS84_SRID))
    utm_srid = Column(Integer)
    # used to detect folding back
    route_a = Column(Geography('LineStringM', srid=WGS84_SRID))  # delta azimuth
    # used to calculate values from offsets
    route_d = Column(Geography('LineStringM', srid=WGS84_SRID))  # distance
    route_et = Column(Geography('LineStringZM', srid=WGS84_SRID))  # time
    # used to detect climbs (cannot be substringed)
    route_edt = Column(Geography('LineStringZM', srid=WGS84_SRID))  # elevation, distance / m * 1e7 + elapsed time

    __mapper_args__ = {
        'polymorphic_identity': SourceType.ACTIVITY
    }

    def __str__(self):
        return 'ActivityJournal %s %s to %s' % (self.activity_group.name,
                                                format_time(self.start), format_time(self.finish))

    def time_range(self, s):
        return self.start, self.finish

    def get_all_qname(self, s, qname, limit=True):
        direct = self.get_qname(s, qname, limit=limit)
        indirect = self.get_activity_topic_journal(s).get_qname(s, qname, limit=limit)
        log.debug(f'{len(direct)} {len(indirect)}')
        return direct + indirect

    def get_activity_topic_journal(self, s):
        from .. import ActivityTopicJournal, FileHash
        with timing('get_activity_topic_journal'):
            return s.query(ActivityTopicJournal). \
                join(FileHash). \
                join(ActivityJournal). \
                filter(ActivityJournal.id == self.id).one()

    @classmethod
    def at(cls, s, local_time_or_date, activity_group=None):
        try:
            activity_journal = cls.at_local_time(s, local_time_or_date, activity_group=activity_group)
        except Exception as e:
            log.debug(e)
            activity_journal = cls.at_date(s, local_time_or_date, activity_group=activity_group)
        if activity_group and activity_journal.activity_group != ActivityGroup.from_name(s, activity_group):
            raise Exception(f'Activity journal from {local_time_or_date} '
                            f'does not match activity group {activity_group}')
        log.info(f'Resolved {local_time_or_date} / {activity_group} to {activity_journal}')
        return activity_journal

    @classmethod
    def at_date(cls, s, date, activity_group=None):
        day = local_date_to_time(date)
        journals = s.query(ActivityJournal).filter(ActivityJournal.start >= day,
                                                   ActivityJournal.start < day + dt.timedelta(days=1)).all()
        if activity_group:
            activity_group = ActivityGroup.from_name(s, activity_group)
            journals = [journal for journal in journals if journal.activity_group == activity_group]
        if not journals:
            msg = f'No activity journal found at {date}'
            if activity_group: msg += f' for {activity_group}'
            raise Exception(msg)
        elif len(journals) > 1:
            msg = f'Multiple activity journals found at {date}'
            if activity_group: msg += f' for {activity_group}'
            raise Exception(msg)
        else:
            return journals[0]

    @classmethod
    def at_local_time(cls, s, local_time, activity_group=None):
        time = local_time_to_time(local_time)
        q = s.query(ActivityJournal).filter(ActivityJournal.start == time)
        if activity_group:
            activity_group = ActivityGroup.from_name(s, activity_group)
            q = q.filter(ActivityJournal.activity_group_id == activity_group.id)
        if q.count():
            return q.one()
        else:
            msg = f'No activity found at {local_time} ({time})'
            if activity_group: msg += f' for {activity_group}'
            raise Exception(msg)

    @classmethod
    def before_local_time(cls, s, local_time):
        time = local_time_to_time(local_time)
        return s.query(ActivityJournal). \
            filter(ActivityJournal.start < time). \
            order_by(desc(ActivityJournal.start)). \
            limit(1).one_or_none()

    @classmethod
    def after_local_time(cls, s, local_time):
        time = local_time_to_time(local_time)
        return s.query(ActivityJournal). \
            filter(ActivityJournal.start >= (time + dt.timedelta(days=1))). \
            order_by(ActivityJournal.start). \
            limit(1).one_or_none()

    @classmethod
    def number_of_activities(cls, s):
        return s.query(ActivityJournal).count()

    def __lt__(self, other):
        # allows sorting in displays
        if isinstance(other, ActivityJournal):
            return self.start < other.start
        else:
            return False


class ActivityTimespan(Base):

    __tablename__ = 'activity_timespan'

    id = Column(Integer, primary_key=True)
    activity_journal_id = Column(Integer, ForeignKey('source.id', ondelete='cascade'),
                                 nullable=False, index=True)
    activity_journal = relationship('ActivityJournal',
                                    backref=backref('timespans', cascade='all, delete-orphan',
                                                    passive_deletes=True,
                                                    order_by='ActivityTimespan.start'))
    start = Column(UTC, nullable=False)
    finish = Column(UTC, nullable=False)
    UniqueConstraint(activity_journal_id, start)

    def __str__(self):
        return 'ActivityTimespan from %s - %s' % (format_time(self.start), format_time(self.finish))


class ActivityBookmark(Base):

    __tablename__ = 'activity_bookmark'

    id = Column(Integer, primary_key=True)
    activity_journal_id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), nullable=False)
    activity_journal = relationship('ActivityJournal')
    start = Column(UTC, nullable=False)
    finish = Column(UTC, nullable=False)
    owner = Column(ShortCls, nullable=False, index=True)  # index for deletion
    constraint = Column(NullText, index=True)
    UniqueConstraint(activity_journal_id, start, finish, owner, constraint)

    def __str__(self):
        return 'ActivityBookmark from %s - %s' % (format_time(self.start), format_time(self.finish))





from collections import namedtuple
from json import loads
from logging import getLogger

from .utils import ProcessCalculator, ActivityGroupProcessCalculator, DataFrameCalculatorMixin
from ..pipeline import OwnerInMixin, LoaderMixin
from ...common.math import is_nan
from ...data import Statistics
from ...data.impulse import hr_zone, impulse_10
from ...names import N, T, SPACE
from ...sql import Constant, StatisticJournalType

log = getLogger(__name__)

# constraint comes from constant
HRImpulse = namedtuple('HRImpulse', 'title, gamma, zero, one, max_secs')


class ImpulseCalculator(LoaderMixin, OwnerInMixin, DataFrameCalculatorMixin,
                        ActivityGroupProcessCalculator):

    def __init__(self, *args, prefix=None, impulse_constant=None, activity_group=None, **kargs):
        self.impulse_constant_ref = self._assert('impulse_constant', impulse_constant)
        self.prefix = self._assert('prefix', prefix)
        super().__init__(*args, activity_group=activity_group, **kargs)

    def _startup(self, s):
        super()._startup(s)
        self.impulse_constant = Constant.from_name(s, self.impulse_constant_ref)
        self.impulse = HRImpulse(**loads(self.impulse_constant.at(s).value))
        log.debug('%s: %s' % (self.impulse_constant, self.impulse))
        self._provides(s, T.HR_ZONE, StatisticJournalType.FLOAT, None, None,
                       'The SHRIMP HR zone.')
        name_group = self.prefix + SPACE + self.impulse_constant.short_name
        self._provides(s, name_group, StatisticJournalType.FLOAT, None, None,
                       'The SHRIMP HR impulse over 10 seconds.', title=self.impulse.title)

    def _read_dataframe(self, s, ajournal):
        try:
            heart_rate_df = Statistics(s, activity_journal=ajournal). \
                by_name(self.owner_in, N.HEART_RATE).df
            fthr_df = Statistics(s).by_name(Constant, N.FTHR).df
        except Exception as e:
            log.warning(f'Failed to generate statistics for activity: {e}')
            raise
        if fthr_df.empty:
            raise Exception(f'No {N.FTHR} defined for {ajournal.activity_group}')
        return heart_rate_df, fthr_df

    def _calculate_stats(self, s, ajournal, data):
        heart_rate_df, fthr_df = data
        hr_zone(heart_rate_df, fthr_df)
        impulse_df = impulse_10(heart_rate_df, self.impulse)
        # join so that we can iterate over values in time order
        stats = impulse_df.join(heart_rate_df, how='outer')
        return stats

    def _copy_results(self, s, ajournal, loader, stats):
        name_group = self.prefix + SPACE + self.impulse_constant.short_name  # drop activity group as present elsewhere
        for time, row in stats.iterrows():
            if N.HR_ZONE in row and not is_nan(row[N.HR_ZONE]):
                loader.add_data(N.HR_ZONE, ajournal, row[N.HR_ZONE], time)
            if N.HR_IMPULSE_10 in row and not is_nan(row[N.HR_IMPULSE_10]):
                loader.add_data(name_group, ajournal, row[N.HR_IMPULSE_10], time)
        # if there are no values, add a single 1 so we don't re-process
        if not loader:
            loader.add_data(N.HR_ZONE, ajournal, 1, ajournal.start)




from logging import getLogger
from os.path import join, exists

from matplotlib import use
from matplotlib.pyplot import show, figure
from sqlalchemy import alias

from .args import ACTIVITY, IMAGE_DIR, DISPLAY, STATISTIC, SPARKLINE, SECTOR, INVERT
from ..common.log import log_current_exception
from ..common.plot import new_fig, new_ax, normalize, ORANGE
from ..sql import StatisticJournal, Sector, ActivityJournal
from ..sql.tables.sector import SectorJournal

log = getLogger(__name__)


def sparkline(config):
    '''
## sparkline

    > ch2 sparkline SECTOR-ID

Generate a sparkline plot of the statistic.
    '''
    with config.db.session_context() as s:
        if config.args[DISPLAY]:
            display(s, config.args[STATISTIC], config.args[SECTOR], config.args[ACTIVITY], config.args[INVERT])
        else:
            create_in_cache(config.args._format_path(IMAGE_DIR), s,
                            config.args[STATISTIC], config.args[SECTOR], config.args[ACTIVITY], config.args[INVERT])


def read_statistic(s, statistic_id, sector_id, activity_id):
    try:
        if sector_id:
            instances = s.query(StatisticJournal, ActivityJournal). \
                join(SectorJournal, SectorJournal.id == StatisticJournal.source_id). \
                join(Sector, Sector.id == SectorJournal.sector_id). \
                join(ActivityJournal, ActivityJournal.id == SectorJournal.activity_journal_id). \
                filter(Sector.id == sector_id,
                       StatisticJournal.statistic_name_id == statistic_id). \
                order_by(StatisticJournal.time).all()
        else:
            instances = s.query(StatisticJournal, ActivityJournal). \
                filter(ActivityJournal.id == StatisticJournal.source_id). \
                filter(StatisticJournal.statistic_name_id == statistic_id). \
                order_by(StatisticJournal.time).all()
        if activity_id:
            activity_journal = ActivityJournal.from_id(s, activity_id)
            data = [(instance[0].time, instance[0].value, instance[1].id == activity_id)
                    for instance in instances if instance[1].activity_group_id == activity_journal.activity_group_id]
        else:
            data = [(instance[0].time, instance[0].value, False) for instance in instances]
        if data:
            # we have no x units - just need a int/float value to scale
            t0 = data[0][0]
            data = [((datum[0] - t0).total_seconds(), datum[1], datum[2]) for datum in data]
            return list(zip(*data))
        else:
            return [], [], []
    except:
        log_current_exception()
        raise Exception(f'{statistic_id} is not a valid statistic name ID')


def fig_from_data(data, cm=1, width=7, invert=False):
    fig = new_fig(cm=cm, width=width)
    ax, _ = new_ax(fig, width=width)
    xs, ys, _ = data
    if xs:
        if invert: ys = [1/y for y in ys]
        fx, fy, _, _ = normalize(xs, ys, preserve_aspect_ratio=False)
        xs, ys = [fx(x) for x in xs], [fy(y) for y in ys]
        ax.plot(xs, ys, color='grey')
        ax.plot(xs, ys, marker='o', color='white', linewidth=0, markersize=cm)
        for (x, y, activity) in zip(*data):
            if invert: y = 1 / y
            if activity: ax.plot([fx(x)], [fy(y)], marker='o', color=ORANGE, markersize=cm*2)
    return fig


def display(s, statistic_id, sector_id, activity_id, invert=False):
    data = read_statistic(s, statistic_id, sector_id, activity_id)
    use('tkagg')
    fig = fig_from_data(data, invert=invert)
    fig.gca().set_facecolor('black')
    show()


def create_in_cache(dir, s, statistic_id, sector_id, activity_id, invert=False):
    path = join(dir, f'{SPARKLINE}-{statistic_id}:{sector_id}:{activity_id}:{invert}.png')
    if not exists(path):
        data = read_statistic(s, statistic_id, sector_id, activity_id)
        use('agg')
        fig = fig_from_data(data, invert=invert)
        fig.savefig(path, transparent=True)
    log.info(f'Sparkline in {path}')
    return path




from logging import getLogger

from ..profile import Profile

log = getLogger(__name__)


def default(config):
    '''
## default

The default configuration with basic activity groups, diary topics and FF parameters.
    '''
    Profile(config).load()



import datetime as dt
from abc import abstractmethod
from enum import IntEnum
from logging import getLogger
from operator import or_

from sqlalchemy import ForeignKey, Column, Integer, func, UniqueConstraint, Boolean, Date
from sqlalchemy.event import listens_for
from sqlalchemy.orm import Session, relationship
from sqlalchemy.sql.functions import count

from ..support import Base
from ..triggers import add_child_ddl
from ..types import OpenSched, ShortCls, short_cls
from ..utils import add
from ...common.date import time_to_local_date, max_time, min_time, extend_range
from ...common.names import UNDEF, TIME_ZERO
from ...lib.utils import timing

log = getLogger(__name__)


class SourceType(IntEnum):

    SOURCE = 0
    INTERVAL = 1
    ACTIVITY = 2
    DIARY_TOPIC = 3
    CONSTANT = 4
    MONITOR = 5
    SEGMENT = 6
    COMPOSITE = 7
    ITEM = 8
    MODEL = 9
    ACTIVITY_TOPIC = 10
    SECTOR = 11


class Source(Base):

    __tablename__ = 'source'

    id = Column(Integer, primary_key=True)
    type = Column(Integer, nullable=False, index=True)  # index needed for fast deletes of subtypes
    activity_group_id = Column(Integer, ForeignKey('activity_group.id', ondelete='cascade'), nullable=True)
    activity_group = relationship('ActivityGroup')

    __mapper_args__ = {
        'polymorphic_identity': SourceType.SOURCE,
        'polymorphic_on': type
    }

    @abstractmethod
    def time_range(self, s):
        raise NotImplementedError('time_range for %s' % self)

    @classmethod
    def before_flush(cls, s):
        cls.__record_dirty_intervals(s)

    @classmethod
    def __record_dirty_intervals(cls, s):
        from .statistic import StatisticJournal, StatisticJournalText
        # sessions are generally restricted to one time region, so we'll bracket that rather than list all times
        start, finish = None, None
        # all sources except intervals that are being deleted (need to catch on cascade to statistics)
        for instance in s.deleted:
            if isinstance(instance, Source) and not isinstance(instance, Interval):
                a, b = instance.time_range(s)
                start, finish = min_time(a, start), max_time(b, finish)
        # all modified statistics
        for instance in s.dirty:
            # ignore constants as time 0
            # ignore textual values (don't have useful stats and commonly modified fields)
            if isinstance(instance, StatisticJournal):
                if s.is_modified(instance) and instance.time and not isinstance(instance, StatisticJournalText):
                    start, finish = extend_range(start, finish, instance.time)
                elif instance.statistic_name is None or instance.source is None:
                    log.warning(f'Loading with incomplete name/source data (bad dirty logic) '
                                f'(statistic_name_id {instance.statistic_name_id})')
        # all new statistics that aren't associated with intervals and have non-null data
        # (avoid triggering on empty diary entries)
        for instance in s.new:
            # ignore constants as time 0
            # ignore textual values
            if isinstance(instance, StatisticJournal):
                if not isinstance(instance.source, Interval) and not isinstance(instance.source, StatisticJournalText):
                    start, finish = extend_range(start, finish, instance.time)
                elif instance.statistic_name is None or instance.source is None:
                    log.warning(f'Loading with incomplete name/source data (bad dirty logic) '
                                f'(statistic_name_id {instance.statistic_name_id})')
        if start is not None:
            Interval.record_dirty_times(s, start, finish)

    @classmethod
    def from_id(cls, s, id):
        return s.query(Source).filter(Source.id == id).one()

    def long_str(self):
        return str(self)

    def get_qname(self, s, qname, limit=True):
        from .. import StatisticJournal, StatisticName, ActivityGroup
        owner, name, group = StatisticName.parse(qname, default_activity_group=UNDEF)
        q = s.query(StatisticJournal). \
            join(StatisticName). \
            filter(StatisticName.name.ilike(name),
                   StatisticJournal.source_id == self.id)
        if owner:
            q = q.filter(StatisticName.owner == owner)
        if group is not UNDEF:
            q = q.join(ActivityGroup).filter(ActivityGroup.name == group)
        if limit:
            q = q.filter(or_(StatisticJournal.serial.is_(None), StatisticJournal.serial == 0))
        log.debug(q)
        with timing(f'get_qname {self.id} {owner} {name} {group}'):
            return q.all()



@listens_for(Session, 'before_flush')
def before_flush(session, context, instances):
    from .. import StatisticJournal
    StatisticJournal.before_flush(session)
    Source.before_flush(session)


class GroupedSource(Source):

    __abstract__ = True

    def __init__(self, **kargs):
        if 'activity_group' not in kargs: raise Exception(f'{self.__class__.__name__} requires activity group')
        super().__init__(**kargs)

    def long_str(self):
        return super().long_str() + f' ({self.activity_group.name})'


class UngroupedSource(Source):

    __abstract__ = True

    def __init__(self, **kargs):
        if 'activity_group' in kargs: raise Exception(f'{self.__class__.__name__} has no activity group')
        super().__init__(**kargs)


class NoStatistics(Exception):
    pass


@add_child_ddl(Source)
class Interval(Source):

    __tablename__ = 'interval'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    schedule = Column(OpenSched, nullable=False, index=True)
    owner = Column(ShortCls, nullable=False)
    dirty = Column(Boolean, default=False, nullable=False)
    permanent = Column(Boolean, default=False, nullable=False)  # if true, do not clean when dirty
    # these are for the schedule - finish is redundant (start is not because of timezone issues)
    start = Column(Date, nullable=False, index=True)
    finish = Column(Date, nullable=False, index=True)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.INTERVAL
    }

    def __init__(self, start=None, finish=None, schedule=None, **kargs):
        if not start: raise ValueError('Missing start')
        if not schedule: raise ValueError('Missing schedule')
        if not finish: finish = schedule.next_frame(start)
        super().__init__(start=start, finish=finish, schedule=schedule, **kargs)

    def __str__(self):
        owner = self.owner if isinstance(self.owner, str) else short_cls(self.owner)
        return 'Interval "%s from %s" (owner %s)' % (self.schedule, self.start, owner)

    @classmethod
    def _raw_statistics_time_range(cls, s, exclude_owners=None):
        '''
        The time range over which statistics exist, ignoring constants at "time zero".
        This is the first to the last time for any statistics - it pays no attention to gaps.
        '''
        from .statistic import StatisticJournal, StatisticName
        # exclude 2 days from zero because here are some stats set at the date equivalent to zero time,
        # but because of date/time conversions and timezones these are not exactly as time zero
        q = s.query(func.min(StatisticJournal.time), func.max(StatisticJournal.time)). \
            filter(StatisticJournal.time > TIME_ZERO + dt.timedelta(days=2))
        if exclude_owners:
            q = q.join(StatisticName).filter(~StatisticName.owner.in_(exclude_owners))
        start, finish = q.one()
        if start and finish:
            return start, finish
        else:
            raise NoStatistics('No statistics are currently defined')

    @classmethod
    def missing_starts(cls, s, expected, schedule, interval_owner, exclude_owners=None):
        '''
        Previous approach was way too complicated and not thread-safe.  Instead, just enumerate intervals and test.
        '''
        try:
            stats_start_time, stats_finish_time = cls._raw_statistics_time_range(s, exclude_owners=exclude_owners)
            start = time_to_local_date(stats_start_time)
            finish = time_to_local_date(stats_finish_time)
            log.debug('Statistics (in general) exist %s - %s' % (start, finish))
            start = schedule.start_of_frame(start)
            finish = schedule.next_frame(finish)
            while start < finish:
                existing = s.query(Interval). \
                    filter(Interval.start == start,
                           Interval.schedule == schedule,
                           Interval.owner == interval_owner).count()
                if existing != expected:
                    yield start
                start = schedule.next_frame(start)
        except NoStatistics:
            log.warning('No data available')

    @classmethod
    def dirty_all(cls, s):
        log.warning('Dirtying all Intervals')
        s.record_dirty_intervals(interval.id for interval in s.query(Interval).all())

    @classmethod
    def record_dirty_times(cls, s, start, finish):
        '''
        Record dirty intervals that include data in the given TIME range,
        '''
        start, finish = time_to_local_date(start), time_to_local_date(finish)
        q = s.query(Interval.id).filter(Interval.start <= finish, Interval.finish > start)
        # do not mark in-place because we can get deadlock transactions.
        # instead, save in session and update at end of transaction
        s.record_dirty_intervals(row[0] for row in q.all())

    @classmethod
    def clean(cls, s, owner=None):
        q = s.query(Interval.id).filter(Interval.dirty == True)
        if owner:
            extra = f' (owner {owner})'
            q = q.filter(Interval.owner == owner)
        else:
            extra = ' (not permanent)'
            q = q.filter(Interval.permanent == False)
        count = q.count()
        if count:
            log.debug(f'Cleaning {count} dirty intervals {extra}')
            s.query(Source).filter(Source.id.in_(q)).delete(synchronize_session=False)
            s.commit()
        log.debug('Intervals clean')


class CompositeComponent(Base):

    __tablename__ = 'composite_component'

    id = Column(Integer, primary_key=True)
    input_source_id = Column(Integer, ForeignKey('source.id', ondelete='cascade'),
                             nullable=False, index=True)
    input_source = relationship('Source', foreign_keys=[input_source_id])
    output_source_id = Column(Integer, ForeignKey('composite_source.id', ondelete='cascade'),
                              nullable=False)  # no need for index - see unique below
    output_source = relationship('Composite', foreign_keys=[output_source_id])
    UniqueConstraint(output_source_id, input_source_id)  # ordered so output is a useful index


@add_child_ddl(Source)
class Composite(Source):

    __tablename__ = 'composite_source'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    # viewonly below was in response to:
    # home/andrew/project/choochoo/py/env/lib/python3.9/site-packages/SQLAlchemy-1.4.0b1-py3.9-linux-x86_64.egg/sqlalchemy/orm/relationships.py:3428: SAWarning: relationship 'Composite.components' will copy column composite_source.id to column composite_component.output_source_id, which conflicts with relationship(s): 'CompositeComponent.output_source' (copies composite_source.id to composite_component.output_source_id). If this is not the intention, consider if these relationships should be linked with back_populates, or if viewonly=True should be applied to one or more if they are read-only. For the less common case that foreign key constraints are partially overlapping, the orm.foreign() annotation can be used to isolate the columns that should be written towards.   The 'overlaps' parameter may be used to remove this warning.
    # /home/andrew/project/choochoo/py/env/lib/python3.9/site-packages/SQLAlchemy-1.4.0b1-py3.9-linux-x86_64.egg/sqlalchemy/orm/relationships.py:3428: SAWarning: relationship 'Composite.components' will copy column source.id to column composite_component.input_source_id, which conflicts with relationship(s): 'CompositeComponent.input_source' (copies source.id to composite_component.input_source_id). If this is not the intention, consider if these relationships should be linked with back_populates, or if viewonly=True should be applied to one or more if they are read-only. For the less common case that foreign key constraints are partially overlapping, the orm.foreign() annotation can be used to isolate the columns that should be written towards.   The 'overlaps' parameter may be used to remove this warning.
    components = relationship('Source', secondary='composite_component', viewonly=True)
    n_components = Column(Integer, nullable=False)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.COMPOSITE
    }

    def time_range(self, s):
        return None, None

    @classmethod
    def clean(cls, s):
        log.debug('Searching for invalid composites')
        # see test_recursive
        q_input_counts = s.query(Composite.id,
                                 count(CompositeComponent.input_source_id).label('count')). \
            outerjoin(CompositeComponent, CompositeComponent.output_source_id == Composite.id). \
            group_by(Composite.id).cte()
        q_bad_nodes = s.query(Composite.id). \
            join(q_input_counts, q_input_counts.c.id == Composite.id). \
            filter(Composite.n_components != q_input_counts.c.count)
        q_count = s.query(count(Composite.id)).filter(Composite.id.in_(q_bad_nodes))
        # log.debug(q_count)
        if q_count.scalar():
            log.warning('Need to clean expired composite sources (may take some time)')
            q_bad_nodes = q_bad_nodes.cte(recursive=True)
            q_all_nodes = q_bad_nodes. \
                union_all(s.query(Composite.id).
                          join(CompositeComponent,
                               CompositeComponent.output_source_id == Composite.id).
                          join(q_bad_nodes,
                               CompositeComponent.input_source_id == q_bad_nodes.c.id)).select()
            log.debug(f'Executing {q_all_nodes}')
            s.flush()
            with timing('GC of composite sources'):
                s.query(Source).filter(Source.id.in_(q_all_nodes)).delete(synchronize_session=False)

    @classmethod
    def create(cls, s, *sources):
        composite = add(s, Composite(n_components=0))
        for source in sources:
            add(s, CompositeComponent(input_source=source, output_source=composite))
            composite.n_components += 1
        return composite

    def long_str(self):
        interesting = [component for component in self.components if not isinstance(component, Composite)]
        msg = f'Composite (id {self.id}, {self.n_components} comp)'
        if self.activity_group: msg += f' ({self.activity_group.name})'
        if interesting:
            msg += ' [' + ', '.join(source.long_str() for source in interesting) + ']'
        return msg



from collections import defaultdict
from itertools import groupby
from logging import getLogger
from weakref import WeakSet

from sqlalchemy import Sequence, select, text, inspect
from sqlalchemy.event import listen

log = getLogger(__name__)


class BatchLoader:

    def __init__(self, enabled=True, max_msg_cnt=10):
        self.enabled = enabled
        self.__max_msg_cnt = max_msg_cnt
        self.__sessions = WeakSet()
        self.reset()

    def reset(self):
        self.msgs = defaultdict(lambda: 0)
        self.sessions = 0
        self.calls = 0
        self.rows = 0

    def debug(self, msg):
        self.__message('debug', msg, log.debug)

    def warning(self, msg):
        self.__message('warning', msg, log.warning)

    def error(self, msg):
        self.__message('error', msg, log.error)

    def __message(self, type_, msg, log):
        self.msgs[(type_, msg)] += 1
        if self.msgs[(type_, msg)] <= self.__max_msg_cnt:
            log(msg)

    @staticmethod
    def __group_by_mapper(instances):
        sort_key = lambda model: type(model).__mapper__.base_mapper
        return groupby(sorted(instances, key=lambda x: str(sort_key(x))), key=sort_key)

    def __key_ok(self, mapper):
        if len(mapper.primary_key) == 1:
            key = mapper.primary_key[0]
            try:
                if key.type.python_type == int and \
                        key.autoincrement in ('auto', True) and \
                        key.table == mapper.local_table:
                    return True
                else:
                    self.warning(f'Unexpected key type for {mapper}')
            except Exception as e:
                # only debug here because we know this occurs with user-defined types
                # (like system constant)
                self.debug(f'Batch error with key {key}: {e!r}')
        else:
            self.warning(f'Composite primary key for {mapper}')
        return False

    @staticmethod
    def __ids(session, mapper, column, n):
        id_seq_name = f'{mapper.entity.__tablename__}_{column}_seq'
        schema = mapper.entity.__table__.metadata.schema
        sequence = Sequence(id_seq_name, schema=schema)
        return [int(row[0]) for row in session.connection().execute(
            select([sequence.next_value()]).select_from(text("generate_series(1, :num_values)")),
            num_values=n)]

    def __set_ids(self, session, mapper, column, missing):
        n = len(missing)
        for id, instance in zip(self.__ids(session, mapper, column, n), missing):
            setattr(instance, column, id)
        self.rows += 1

    def __populate(self, session):
        for mapper, instances in self.__group_by_mapper(session.new):
            if self.__key_ok(mapper):
                column = mapper.primary_key[0].name
                missing = [instance for instance in instances if getattr(instance, column) is None]
                if missing:
                    self.__set_ids(session, mapper, column, missing)
                else:
                    self.warning(f'ID already populated for {type(instances[0])}')

    @staticmethod
    def __group(session):
        instances = session.new
        sort_key = lambda instance: (str(type(instance)), inspect(instance).insert_order)
        for i, instance in enumerate(sorted(instances, key=sort_key)):
            inspect(instance).insert_order = i

    def enable(self, session):
        if self.enabled:
            if session not in self.__sessions:
                self.__sessions.add(session)
                self.sessions += 1
                listen(session, 'before_flush', self.__before_flush)
            else:
                log.warning(f'Tried to register session more than once')

    def __before_flush(self, session, context, instances):
        try:
            self.calls += 1
            if instances:
                self.warning('No batch support for explicit instances')
            else:
                self.__populate(session)
                self.__group(session)
        except Exception as e:
            self.error(e)
            raise



from itertools import groupby
from logging import getLogger

from sqlalchemy import text
from sqlalchemy.exc import DataError

from ch2.names import N
from ch2.sql import StatisticJournal, StatisticName, Source
from ch2.sql.tables.sector import SectorJournal
from ch2.sql.utils import add

log = getLogger(__name__)


HULL_RADIUS = 25  # 10 and 15 were too small - climbs did not overlap


def find_and_add_sector_journals(s, sector_group, ajournal, sector_id=None):
    sql = text('''
with srid as (select s.id as sector_id,
                     st_setsrid(s.route, sg.srid) as sector,
                     st_transform(aj.route_et::geometry, sg.srid) as route_et,
                     st_transform(aj.route_d::geometry, sg.srid) as route_d,
                     st_setsrid(s.start, sg.srid) as start,
                     st_setsrid(s.finish, sg.srid) as finish
                from sector as s,
                     activity_journal as aj,
                     sector_group as sg
               where s.sector_group_id = sg.id
                 and s.id = coalesce(:sector_id, s.id)
                 and sg.id = :sector_group_id
                 and aj.id = :activity_journal_id
                 and st_intersects(st_setsrid(s.hull, sg.srid), st_transform(aj.route_d::geometry, sg.srid))),
     start_point as (select r.sector_id,
                            r.route_et,
                            (st_dump(st_multi(st_intersection(r.start, st_force2d(r.route_et))))).geom as point
                       from srid as r),
     start_fraction as (select p.sector_id,
                               st_linelocatepoint(p.route_et, p.point) as fraction
                          from start_point as p
                         where st_geometrytype(p.point) = 'ST_Point'),  -- small number of cases intersect as lines
     finish_point as (select r.sector_id,
                             r.route_et,
                             (st_dump(st_multi(st_intersection(r.finish, st_force2d(r.route_et))))).geom as point
                        from srid as r),
     finish_fraction as (select p.sector_id,
                                st_linelocatepoint(p.route_et, p.point) as fraction
                           from finish_point as p
                          where st_geometrytype(p.point) = 'ST_Point'),
     shortest as (select distinct  -- multiple starts/finishes can lead to duplicates
                         r.sector_id,
                         s.fraction as start_fraction,
                         f.fraction as finish_fraction,
                         min(f.fraction - s.fraction) over (partition by r.sector_id) as shortest
                    from srid as r,
                         start_fraction as s,
                         finish_fraction as f
                   where s.fraction < f.fraction
                     and s.sector_id = f.sector_id
                     and s.sector_id = r.sector_id
                     and st_length(st_linesubstring(r.route_d, s.fraction, f.fraction))
                         between 0.95 * st_length(r.sector) and 1.05 * st_length(r.sector))
select s.sector_id,
       s.start_fraction,
       s.finish_fraction,
       aj.start + interval '1' second * st_m(st_lineinterpolatepoint(r.route_et, s.start_fraction)) as start_time,
       aj.start + interval '1' second * st_m(st_lineinterpolatepoint(r.route_et, s.finish_fraction)) as finish_time,
       st_m(st_lineinterpolatepoint(r.route_d, s.start_fraction)) as start_distance,
       st_m(st_lineinterpolatepoint(r.route_d, s.finish_fraction)) as finish_distance,
       st_z(st_lineinterpolatepoint(r.route_et, s.start_fraction)) as start_elevation,
       st_z(st_lineinterpolatepoint(r.route_et, s.finish_fraction)) as finish_elevation
  from srid as r,
       shortest as s,
       activity_journal as aj
 where aj.id = :activity_journal_id
   and r.sector_id = s.sector_id
   and s.finish_fraction - s.start_fraction = s.shortest
''')
    log.debug(sql)
    result = s.connection().execute(sql, sector_group_id=sector_group.id, activity_journal_id=ajournal.id,
                                    sector_id=sector_id)
    for row in result.fetchall():
        data = {name: value for name, value in zip(result.keys(), row)}
        log.debug(f'Adding SectorJournal for activity_journal_id {ajournal.id}, '
                  f'activity_group {ajournal.activity_group}: {data}')
        sjournal = add(s, SectorJournal(activity_journal_id=ajournal.id, activity_group=ajournal.activity_group,
                                        **data))
        s.flush()
        yield sjournal


def add_sector_statistics(s, sjournal, loader, **kargs):
    # delegate to the sector since that can be subclassed (eg climb)
    sjournal.sector.add_statistics(s, sjournal, loader, **kargs)


def add_start_finish(s, sector_id):
    # need to take just the final metre at each end of the route and shift to avoid multilines
    # with 'jigsaw' shaped routes (endpoint of multiline is null).
    sql = text('''
with srid as (select st_setsrid(s.route, sg.srid) as route,
                     st_length(st_setsrid(s.route, sg.srid)) as length
                from sector as s,
                     sector_group as sg
               where s.id = :sector_id
                 and sg.id = s.sector_group_id),
     ends as (select st_linesubstring(route, 0, 1/length) as start,
                     st_linesubstring(route, 1-1/length, 1) as finish,
                     route
                from srid),
     offsets as (select st_offsetcurve(start, :radius) as start_left,
                        st_offsetcurve(start, -:radius) as start_right,
                        st_offsetcurve(finish, :radius) as finish_left,
                        st_offsetcurve(finish, -:radius) as finish_right,
                        st_buffer(route, :radius, 'endcap=flat') as hull,
                        route
                   from ends),
     endcaps as (select st_makeline(st_startpoint(start_left), st_endpoint(start_right)) as start,
                        st_makeline(st_endpoint(finish_left), st_startpoint(finish_right)) as finish,
                        o.hull,
                        o.route
                   from offsets as o)
update sector
   set start = e.start,
       finish = e.finish,
       hull = e.hull,
       distance = coalesce(distance, st_length(e.route) / 1000)
  from endcaps as e
 where sector.id = :sector_id
   and st_geometrytype(e.hull) = 'ST_Polygon'
''')
    log.debug(sql)
    s.connection().execute(sql, sector_id=sector_id, radius=HULL_RADIUS)


def sector_stats_for_activity(s, ajournal):

    from ..pipeline.calculate.sector import SectorCalculator

    query = s.query(StatisticJournal). \
        join(StatisticName).join(Source). \
        join(SectorJournal, SectorJournal.id == Source.id). \
        filter(StatisticName.name.like(N.SECTOR_ANY),
               StatisticJournal.time >= ajournal.start,
               StatisticJournal.time <= ajournal.finish,
               StatisticName.owner == SectorCalculator,
               Source.activity_group == ajournal.activity_group).order_by(StatisticJournal.time)
    sjournals = query.all()

    def make_stats(sjournals):
        sjournals = list(sjournals)
        stats = {sjournal.statistic_name.name: sjournal for sjournal in sjournals}
        stats['sector-journal'] = sjournals[0].source
        return stats

    return sorted((make_stats(grouped)
                   for _, grouped in groupby(sjournals, key=lambda sjournal: sjournal.time)),
                   key=lambda stat: stat['sector-journal'].start_distance)




from collections import namedtuple
from json import loads
from logging import getLogger

import numpy as np
import pandas as pd

from .utils import ActivityGroupProcessCalculator, DataFrameCalculatorMixin, ProcessCalculator
from ..pipeline import LoaderMixin
from ...common.log import log_current_exception
from ...data import present, linear_resample_time, Statistics
from ...data.frame import median_dt
from ...data.lib import interpolate_to_index
from ...data.power import add_differentials, add_energy_budget, add_loss_estimate, add_power_estimate
from ...lib.data import reftuple, MissingReference
from ...names import N, U, S, T
from ...sql import Constant, StatisticJournalType

log = getLogger(__name__)

BikeModel = namedtuple('BikeModel', 'cda, crr, bike_weight')


class PowerModel(reftuple('Power', 'bike_model, rider_weight')):

    def expand(self, s, time, default_owner=None, default_activity_group=None):
        instance = super().expand(s, time, default_owner=default_owner, default_activity_group=default_activity_group)
        bike_model = instance.bike_model
        if not isinstance(bike_model, BikeModel):
            return instance._replace(bike_model=BikeModel(**bike_model))
        else:
            return instance


class PowerCalculator(LoaderMixin, DataFrameCalculatorMixin, ActivityGroupProcessCalculator):

    def __init__(self, *args, power_model=None, caloric_eff=0.25, activity_group=None, **kargs):
        self.power_model_ref = power_model
        self.caloric_eff = caloric_eff
        super().__init__(*args, activity_group=activity_group, **kargs)

    def _startup(self, s):
        super()._startup(s)
        self._provides(s, T.POWER_ESTIMATE, StatisticJournalType.FLOAT, U.W, S.AVG,
                       'The estimated power.')
        self._provides(s, T.VERTICAL_POWER, StatisticJournalType.FLOAT, U.W, S.AVG,
                       'The estimated power from height gain alone.')
        self._provides(s, T.HEADING, StatisticJournalType.FLOAT, U.DEG, None,
                       'The current heading.')
        self._provides(s, T.ENERGY_ESTIMATE, StatisticJournalType.FLOAT, U.KJ, S.MAX,
                       'The estimated total energy expended.')
        self._provides(s, T.CALORIE_ESTIMATE, StatisticJournalType.FLOAT, U.KCAL, S.MAX,
                       'The estimated calories burnt.')

    def _set_power(self, s, ajournal):
        power_model = PowerModel(**loads(Constant.from_name(s, self.power_model_ref).at(s).value))
        self.power_model = power_model.expand(s, ajournal.start,
                                              default_owner=Constant, default_activity_group=self.activity_group)
        log.debug(f'Power: {self.power_model_ref}: {self.power_model}')

    def _read_dataframe(self, s, ajournal):
        from ..owners import ActivityReader, ElevationCalculator
        try:
            self._set_power(s, ajournal)
            df = Statistics(s, activity_journal=ajournal, with_timespan=True). \
                by_name(ActivityReader, N.DISTANCE, N.SPEED, N.CADENCE). \
                by_name(ElevationCalculator, N.ELEVATION).df
            df[N.DISTANCE] = df[N.DISTANCE] * 1000  # convert to SI base units (m) - we're doing PHYSICS
            ldf = linear_resample_time(df)
            ldf = add_differentials(ldf, max_gap=1.1 * median_dt(df))
            return df, ldf
        except MissingReference as e:
            log.warning(f'Power configuration incorrect ({e})')
        except Exception as e:
            log.warning(f'Failed to generate statistics for power ({ajournal.activity_group.name}): {e}')
            log_current_exception()

    def _calculate_stats(self, s, ajournal, dfs):
        df, ldf = dfs
        total_weight = self.power_model.bike_model.bike_weight + self.power_model.rider_weight
        ldf = add_energy_budget(ldf, total_weight)
        ldf = add_loss_estimate(ldf, total_weight, cda=self.power_model.bike_model.cda,
                                crr=self.power_model.bike_model.crr)
        ldf = add_power_estimate(ldf)
        return df, ldf

    def _copy_results(self, s, ajournal, loader, dfs, fields=(N.POWER_ESTIMATE, N.VERTICAL_POWER)):
        df, ldf = dfs
        self.__add_total_energy(s, ajournal, loader, ldf)
        df = interpolate_to_index(df, ldf, *fields)
        for time, row in df.iterrows():
            for name in fields:
                if name in row and not pd.isnull(row[name]):
                    loader.add_data(name, ajournal, row[name], time)

    def __add_total_energy(self, s, ajournal, loader, ldf):
        if present(ldf, N.POWER_ESTIMATE):
            ldf['tmp'] = ldf[N.POWER_ESTIMATE]
            ldf.loc[ldf['tmp'].isna(), ['tmp']] = 0
            energy = np.trapz(y=ldf['tmp'], x=ldf.index.view(np.int64) / 1e12)
            loader.add_data(N.ENERGY_ESTIMATE, ajournal, energy, ajournal.start)
            loader.add_data(N.CALORIE_ESTIMATE, ajournal, energy * 0.239006 / self.caloric_eff, ajournal.start)
            ldf.drop(columns=['tmp'], inplace=True)



from logging import getLogger

from .args import SUB_COMMAND, LIST, PROFILE, ITEM, USERS, SCHEMAS, DATABASES, PROFILES, ADD, DATABASE, SCHEMA, REMOVE, \
    BACKUP
from ..common.db import get_cnxn, add_schema, with_log, remove_schema, remove_database, remove_user, add_database, \
    add_user, list_databases, list_schemas, list_users, backup_schema
from ..common.md import Markdown
from ..common.names import USER, assert_name
from ..config.profile import get_profile, get_profiles
from ..sql.support import Base

log = getLogger(__name__)


def db(config):
    '''
## db

    > ch2 db list users
    > ch2 db list schemas
    > ch2 db list profiles
    > ch2 db list databases

    > ch2 db add user
    > ch2 db add profile PROFILE
    > ch2 db add database

    > ch2 db remove user
    > ch2 db remove schema
    > ch2 db remove database

### Utilities for managing the database.

This command uses the same configuration parameters, and makes the same assumptions about how the system
works, as other commands.  So creating a database creates a database for the current version, adding a
user adds the current user, etc.

Postgres provides a database cluster at a given address/port.  This contains various users and databases.
To store activity data we use a database for each (minor) release version (eg activity-0-35).
Within any version/database, a user can configure a schema with whatever profile they want.
The schema is named after, and owned by, the user and the user can only see data in the schema they own.

There is no 'db add schema' because the schema is added implicitly when the profile is added;
there is no 'db remove profile' because the profile is removed implicitly when the schema is removed
(a schema contains a profile).

### Note On Backups

When a schema is deleted it is copied to `original_name:previous`.  This is done so that user entries
can be read across when re-installing the same version.  That is all.  There is no greater backup
functionality implied or supported.  Backup the database separately if it is important.
'''
    args = config.args
    action, item = args[SUB_COMMAND], args[ITEM]
    {LIST: {USERS: list_users,
            SCHEMAS: list_schemas,
            DATABASES: list_databases,
            PROFILES: list_profiles},
     ADD: {USER: add_user,
           DATABASE: add_database,
           PROFILE: add_profile},
     BACKUP: {SCHEMA: backup_schema},
     REMOVE: {USER: remove_user,
              DATABASE: remove_database,
              SCHEMA: remove_schema}}[action][item](config)


def list_profiles(config):
    fmt = Markdown()
    for name in get_profiles():
        fn, spec = get_profile(name)
        if fn.__doc__:
            fmt.print(fn.__doc__)
        else:
            print(f' ## {name} - lacks docstring\n')


def add_profile(config):
    add_schema(config)
    with with_log('Creating tables'):
        Base.metadata.create_all(config.db.engine)
    profile = config.args[PROFILE]
    fn, spec = get_profile(profile)
    with with_log(f'Loading profile {profile}'):
        fn(config)




from enum import IntEnum
from json import dumps
from logging import getLogger

from sqlalchemy import Column, Integer, not_, or_, Table, ForeignKey, Text
from sqlalchemy.orm import relationship, joinedload

from ..support import Base
from ..types import Cls, Json, Sort, short_cls

log = getLogger(__name__)


class PipelineType(IntEnum):

    PROCESS = 0
    DISPLAY = 1
    DISPLAY_ACTIVITY = 2
    SECTOR = 3


# todo - possibly this should be expressed as a dependency on services, rather than directly between pipelines.
# (i tried this, getting as far as sketching out the tables, and it seemed like a lot of complexity for not
# much gain when the configuration is not varied much - might be more useful in future with a more flexible
# system).

# purely for joins, so not exposed by sqlalchemy
PipelineDependency = Table('pipeline_dependency', Base.metadata,
                           Column('blocks', Integer, ForeignKey('pipeline.id'), primary_key=True),
                           Column('blocked_by', Integer, ForeignKey('pipeline.id'), primary_key=True))


class Pipeline(Base):

    __tablename__ = 'pipeline'

    id = Column(Integer, primary_key=True)
    type = Column(Integer, nullable=False, index=True)
    cls = Column(Cls, nullable=False)  # not unique - may run various instances
    kargs = Column(Json, nullable=False, server_default=dumps({}))

    # https://stackoverflow.com/a/5652169 (no idea why i need both here)
    blocks = relationship('Pipeline', secondary=PipelineDependency,
                          primaryjoin=PipelineDependency.c.blocked_by == id,
                          secondaryjoin=PipelineDependency.c.blocks == id,
                          back_populates='blocked_by')
    blocked_by = relationship('Pipeline', secondary=PipelineDependency,
                              primaryjoin=PipelineDependency.c.blocks == id,
                              secondaryjoin=PipelineDependency.c.blocked_by == id,
                              back_populates='blocks')

    @classmethod
    def _query(cls, s, type=None, like=tuple(), id=None):
        q = s.query(Pipeline)
        if type is not None:  # enum can be 0
            q = q.filter(Pipeline.type == type)
        if like:
            q = q.filter(or_(*[Pipeline.cls.like(pattern) for pattern in like]))
        if id:
            q = q.filter(Pipeline.id == id)
        return q

    @classmethod
    def all(cls, s, type, like=tuple(), id=None):
        pipelines = cls._query(s, type, like=like, id=id).all()
        if not pipelines:
            msg = 'No pipelines configured for type %s' % PipelineType(type).name
            if like:
                msg += f' like {like}'
            if id:
                msg += f' with ID={id}'
            raise Exception(msg)
        return pipelines

    @classmethod
    def all_instances(cls, s, type, like=tuple(), id=None):
        for pipeline in sort_pipelines(cls.all(s, type, like=like, id=id)):
            log.debug(f'Building {pipeline.cls} ({pipeline.kargs})')
            yield pipeline.cls(**pipeline.kargs)

    @classmethod
    def count(cls, s, type, like=tuple(), id=None):
        q = cls._query(s, type, like=like, id=id)
        return q.count()

    def __str__(self):
        return short_cls(self.cls)


def sort_pipelines(pipelines):
    '''
    not only does this order pipelines so that, if run in order, none is blocked, it also expands the
    graph so that when the session is disconnected we have all the data we need.
    '''
    included, processed, remaining = set(pipelines), set(), set(pipelines)
    log.debug(f'Sorting {", ".join(str(pipeline) for pipeline in included)}')
    while remaining:
        for pipeline in remaining:
            if all(blocker not in included or blocker in processed for blocker in pipeline.blocked_by):
                yield pipeline
                processed.add(pipeline)
        remaining = remaining.difference(processed)



from logging import getLogger
from tempfile import TemporaryDirectory

from ch2.commands.args import DEV, V, BASE, bootstrap_db, UPLOAD
from ch2.commands.upload import upload
from ch2.common.args import mm, m
from ch2.config.profiles.default import default
from ch2.diary.database import read_date
from ch2.diary.model import LABEL, VALUE
from ch2.lib import to_date
from tests import LogTestCase, random_test_user

log = getLogger(__name__)


class TestModel(LogTestCase):

    def test_constant(self):
        user = random_test_user()
        with TemporaryDirectory() as f:
            bootstrap_db(user, mm(BASE), f, m(V), '5', mm(DEV), configurator=default)
            config = bootstrap_db(user, mm(BASE), f, mm(DEV), UPLOAD,
                                  'data/test/source/personal/2018-03-04-qdp.fit', '-K', 'n_cpu=1')
            upload(config)
            with config.db.session_context() as s:
                model = list(read_date(s, to_date('2018-03-04')))
                for i, x in enumerate(model):
                    print(i, x)
                [title, diary, shrimp, activity, database] = model
                activity = activity[1][3]  # multiple now supported
                print(activity)
                name = activity[1]
                print(name)
                self.assertEqual(name[LABEL], 'Name')
                self.assertEqual(name[VALUE], '2018-03-04T07:16:33')
                route = activity[3]
                print(route)
                self.assertEqual(route[LABEL], 'Route')



from sqlalchemy import DDL
from sqlalchemy.event import listen


def child_ddl(parent, child, identity):
    '''
    Remove the existing foreign key with constraint and add one without cascade.
    Then re-implement the cascade, but restricting to the correct child table (only).
    Finally, add a warning to catch direct deletion of child tables
    (I did try a reverse cascade, but it causes problems with circular invocation).

    Assumes primary key is called 'id' in both tables, that the discriminator is called 'type',
    and that the existing foreign key constraint is called {child}_id_fkey.

    Note 'before' for child cascades to avoid foreign key conflict.
    '''
    return DDL(f'''
alter table "{child}" drop constraint "{child}_id_fkey";
alter table "{child}" add constraint "{child}_id_fkey"
  foreign key (id) references "{parent}"(id);
create function "{child}_cascade"() returns trigger as $$
  begin
    delete from "{child}" where id = old.id;
    return old;
  end
$$ language plpgsql;
 create trigger "{child}_cascade_trg"
 before delete on "{parent}"
    for each row
   when (old.type = {identity})
execute procedure "{child}_cascade"();
''')


def add_child_ddl(parent_table):

    parent = parent_table.__tablename__

    def decorator(child_table):
        child = child_table.__tablename__
        identity = child_table.__mapper_args__['polymorphic_identity']
        listen(child_table.__table__, 'after_create',
               child_ddl(parent, child, identity).
               execute_if(dialect='postgresql'))
        return child_table

    return decorator


def add_text(ddl):

    def decorator(table_cls):
        listen(table_cls.__table__, 'after_create', DDL(ddl).execute_if(dialect='postgresql'))
        return table_cls

    return decorator




from logging import getLogger
from os.path import exists, join

from matplotlib import use
from matplotlib.pyplot import show, figure

from .args import ACTIVITY, IMAGE_DIR, DISPLAY, SECTOR, THUMBNAIL
from ..common.plot import normalize, new_fig, new_ax, LIME, ORANGE
from ..data.query import Statistics
from ..names import N
from ..pipeline.read.activity import ActivityReader
from ..sql import ActivityJournal, Sector

log = getLogger(__name__)


def thumbnail(config):
    '''
## thumbnail

    > ch2 thumbnail ACTIVITY-ID

Generate a thumbnail map of the activity route.
    '''
    with config.db.session_context() as s:
        if config.args[DISPLAY]:
            display(s, config.args[ACTIVITY], config.args[SECTOR])
        else:
            create_in_cache(config.args._format_path(IMAGE_DIR), s, config.args[ACTIVITY], config.args[SECTOR])


def read_activity(s, activity_id, decimate=10):
    try:
        activity_journal = s.query(ActivityJournal).filter(ActivityJournal.id == activity_id).one()
        df = Statistics(s, activity_journal=activity_journal). \
            by_name(ActivityReader, N.SPHERICAL_MERCATOR_X, N.SPHERICAL_MERCATOR_Y).df
        return df.iloc[::decimate, :]
    except:
        raise Exception(f'{activity_id} is not a valid activity ID')


def read_sector(s, sector_id):
    if sector_id:
        try:
            return s.query(Sector).filter(Sector.id == int(sector_id)).one()
        except:
            raise Exception(f'{sector_id} is not a valid sector ID')


def add_grid(ax, lim, side, grid):
    km = 1000 / side
    n = int(lim / (grid * km)) + 1
    ticks = [km * d * grid for d in range(-n, n+1)]
    ax.set_xticks(ticks)
    ax.set_yticks(ticks)
    ax.grid(axis='both', color='#535353')


def make_figure(xs, ys, side, grid, cm=1.5, border=0.2):
    fig = new_fig(cm=cm)
    ax, lim = new_ax(fig)
    add_grid(ax, lim, side, grid)
    ax.plot(xs, ys, color='white')
    ax.plot([xs[0]], [ys[0]], marker='o', color=LIME, markersize=cm*3)
    ax.plot([xs[-1]], [ys[-1]], marker='o', color=ORANGE, markersize=cm*1.5)
    return fig


def fig_from_df(df, grid=10, cm=1.5, border=0.2):
    xs, ys = zip(*[(x, y) for _, (x, y) in df.iterrows()])
    if xs:
        fx, fy, side, _ = normalize(xs, ys)
        xs, ys = [fx(x) for x in xs], [fy(y) for y in ys]
    else:
        fx, fy = None, None
        xs, ys, side = [0, 0], [0, 0], 1
    return fx, fy, make_figure(xs, ys, side, grid, cm, border)


def display(s, activity_id, sector_id=None):
    df = read_activity(s, activity_id)
    use('tkagg')
    fx, fy, fig = fig_from_df(df)
    if fx and sector_id:
        read_sector(s, sector_id).display(s, fx, fy, fig.gca())
    fig.gca().set_facecolor('black')
    show()


def create_in_cache(dir, s, activity_id, sector_id=None):
    path = join(dir, f'{THUMBNAIL}-{activity_id}:{sector_id if sector_id else None}.png')
    if not exists(path):
        df = read_activity(s, activity_id)
        use('agg')
        fx, fy, fig = fig_from_df(df)
        if fx and sector_id:
            read_sector(s, sector_id).display(s, fx, fy, fig.gca())
        fig.savefig(path, transparent=True)
    log.info(f'Thumbnail in {path}')
    return path



from abc import abstractmethod
from glob import iglob
from logging import getLogger
from os.path import join

from ..pipeline import ProcessPipeline
from ...commands.args import base_system_path, PERMANENT, BASE
from ...common.date import now
from ...common.log import log_current_exception
from ...fit.format.read import filtered_records
from ...lib import to_time
from ...lib.io import modified_file_scans
from ...sql import Timestamp, FileScan

log = getLogger(__name__)


class AbortImport(Exception):
    pass


class AbortImportButMarkScanned(AbortImport):
    pass


class ProcessFitReader(ProcessPipeline):

    def __init__(self, config, *args, sub_dir=None, **kargs):
        self.sub_dir = sub_dir
        super().__init__(config, *args, **kargs)

    def _all_paths(self):
        from ...commands.upload import DOT_FIT
        data_dir = base_system_path(self._config.args[BASE], version=PERMANENT)
        if self.sub_dir:
            data_dir = join(data_dir, self.sub_dir)
        else:
            log.warning('No sub_dir defined - will scan entire tree')
        return iglob(join(data_dir, '**/*' + DOT_FIT), recursive=True)

    def _missing(self, s):
        return [file_scan.path for file_scan in modified_file_scans(s, self._all_paths(), self.owner_out)]

    def _run_one(self, missed):
        with self._config.db.session_context() as s:
            file_scan = s.query(FileScan).filter(FileScan.path == missed).one()
            try:
                self._read(s, file_scan)
                file_scan.last_scan = now()
            except AbortImportButMarkScanned as e:
                log.warning(f'Could not process {file_scan} (scanned)')
                # log_current_exception()
                file_scan.last_scan = now()

    def _read(self, s, file_scan):
        path = file_scan.path  # read this now so that on error we don't go back to database
        try:
            source, data = self._read_data(s, file_scan)
            with Timestamp(owner=self.owner_out, source=source).on_success(s):
                loader = self._get_loader(s, add_serial=True)
                self._load_data(s, loader, data)
                loader.load()
            return loader  # returned so coverage can be accessed
        except Exception as e:
            log_current_exception(traceback=True)
            log.error(f'Error loading {path}: {e}')
            s.rollback()
            raise AbortImportButMarkScanned()

    @staticmethod
    def read_fit_file(data, *options):
        types, messages, records = filtered_records(data)
        return [record.as_dict(*options)
                for _, _, record in sorted(records,
                                           key=lambda r: r[2].timestamp if r[2].timestamp else to_time(0.0))]

    @staticmethod
    def _first(path, records, *names):
        return ProcessFitReader.assert_contained(path, records, names, 0)

    @staticmethod
    def _last(path, records, *names):
        return ProcessFitReader.assert_contained(path, records, names, -1)

    @staticmethod
    def assert_contained(path, records, names, index):
        try:
            return [record for record in records if record.name in names][index]
        except IndexError:
            msg = f'No {names} entry(s) in {path}'
            log.debug(msg)
            raise AbortImportButMarkScanned(msg)

    @abstractmethod
    def _read_data(self, s, file_scan):
        raise NotImplementedError()

    @abstractmethod
    def _load_data(self, s, loader, data):
        raise NotImplementedError()


def quote(text):
    return '"' + text + '"'



from matplotlib.pyplot import figure

ORANGE = '#ff3d00'
LIME = '#cddc39'
CYAN = '#00b8d4'

RED = 'red'
GREEN = 'green'


def stats(zs):
    lo, hi = min(zs), max(zs)
    return lo, hi - lo


def normalize(xs, ys, preserve_aspect_ratio=True):
    xlo, dx = stats(xs)
    ylo, dy = stats(ys)
    if preserve_aspect_ratio:
        if dx > dy:
            ylo -= (dx - dy) / 2
            dy = dx
        else:
            xlo -= (dy - dx) / 2
            dx = dy
    if dx == 0:
        xlo, dx = xlo - 0.5, 1
    if dy == 0:
        ylo, dy = ylo - 0.5, 1
    return lambda x: (x - xlo) / dx - 0.5, lambda y: (y - ylo) / dy - 0.5, dx, dy


def new_fig(cm=1.5, width=1):
    fig = figure(frameon=False)
    fig.set_size_inches(width * cm / 2.54, cm / 2.54)
    return fig


def new_ax(fig, border=0.2, width=1):
    ax = fig.add_subplot(1, 1, 1)
    for edge in ('top', 'right', 'bottom', 'left'):
        ax.spines[edge].set_visible(False)
    lim = 0.5 * (1 + border)
    ax.set_ylim([-lim, lim])
    lim = 0.5 * (1 + border / width)
    ax.set_xlim([-lim, lim])
    ax.tick_params(labelbottom=False, labelleft=False, length=0)
    ax.set_aspect(aspect=1/width, adjustable='box')
    return ax, lim



from json import loads, dumps
from logging import getLogger

from sqlalchemy import exists

from ..worker import run_and_wait
from ...commands.args import DB_VERSION, REMOVE, SCHEMA
from ...commands.db import add_profile
from ...commands.import_ import import_source
from ...common.log import log_current_exception
from ...common.md import HTML, parse, P, LI, PRE, filter_
from ...common.names import BASE, DB, WEB
from ...config.profile import get_profiles
from ...import_ import available_versions
from ...import_.activity import activity_imported
from ...import_.constant import constant_imported
from ...import_.diary import diary_imported
from ...import_.kit import kit_imported
from ...import_.sector import sector_imported
from ...lib import time_to_local_time, local_time_to_time
from ...lib.log import Record
from ...sql import Constant, StatisticJournal, ActivityJournal

log = getLogger(__name__)


ACTIVITY = 'activity'
BUSY = 'busy'
CONFIGURED = 'configured'
CONSTANT = 'constant'
DESCRIPTION = 'description'
DIARY = 'diary'
DIRECTORY = 'directory'
IMPORTED = 'imported'
KIT = 'kit'
NAME = 'name'
PROFILE = 'profile'
PROFILES = 'profiles'
SECTOR = 'sector'
SINGLE = 'single'
STATISTIC = 'statistic'
TIME = 'time'
VALUE = 'value'
VALUES = 'values'
VERSION = 'version'
VERSIONS = 'versions'


class Configure:

    def __init__(self, config):
        self.__config = config
        self.__html = HTML(delta=1, parser=filter_(parse, yes=(P, LI, PRE)))

    def is_configured(self):
        return not bool(self.__config.db.no_data())

    def is_empty(self, s):
        return not s.query(exists().where(ActivityJournal.id > 0)).scalar()

    def is_busy(self):
        try:
            return self.__config.exists_any_process()
        except:  # various, including database having been deleted
            log_current_exception()
            return False

    def html(self, text):
        return self.__html.str(text)

    def read_profiles(self, request, s):
        fn_argspec_by_name = get_profiles()
        data = {PROFILES: {name: self.html(fn_argspec_by_name[name][0].__doc__) for name in fn_argspec_by_name},
                CONFIGURED: not bool(self.__config.db.no_data()),
                BUSY: self.is_busy(),
                VERSION: DB_VERSION,
                DIRECTORY: self.__config.args[BASE]}
        return data

    def write_profile(self, request, s):
        data = request.json
        add_profile(self.__config._with(profile=data[PROFILE]))
        self.__config.reset()

    def delete(self, request, s):
        try:
            # run this as a sub-command so that the user sees that process are running if they try to multi-task
            run_and_wait(self.__config, f'{DB} {REMOVE} {SCHEMA}', Configure, f'{WEB}-{DB}.log')
        except:  # since we delete the database there won't be a process to remove
            log_current_exception()
        self.__config.reset()

    def read_import(self, request, s):
        record = Record(log)
        return {IMPORTED: {DIARY: diary_imported(record, self.__config.db),
                           ACTIVITY: activity_imported(record, self.__config.db),
                           KIT: kit_imported(record, self.__config.db),
                           CONSTANT: constant_imported(record, self.__config.db),
                           SECTOR: sector_imported(record, self.__config.db)},
                VERSIONS: available_versions(self.__config)}

    def write_import(self, request, s):
        data = request.json
        record = Record(log)
        import_source(self.__config, record, data[VERSION])
        return record.json()

    def read_constants(self, request, s):
        return [self.read_constant(s, constant)
                for constant in s.query(Constant).order_by(Constant.name).all()]

    def read_constant(self, s, constant):
        as_json = bool(constant.validate_cls)
        values = [{TIME: time_to_local_time(statistic.time),
                   VALUE: loads(statistic.value) if as_json else statistic.value,
                   STATISTIC: statistic.id}
                  for statistic in s.query(StatisticJournal).
                      filter(StatisticJournal.source_id == constant.id).all()]
        return {NAME: constant.name,
                SINGLE: constant.single,
                DESCRIPTION: self.html(constant.statistic_name.description),
                VALUES: values}

    def write_constant(self, request, s):
        data = request.json
        log.debug(data)
        constant = Constant.from_name(s, data[NAME])
        value = data[VALUES][0][VALUE]
        if constant.validate_cls:
            value = dumps(value)
        time = data[VALUES][0][TIME]
        statistic_journal_id = data[VALUES][0][STATISTIC]
        if statistic_journal_id:
            journal = s.query(StatisticJournal). \
                filter(StatisticJournal.id == statistic_journal_id,
                       StatisticJournal.source_id == constant.id).one()
            journal.value = value
            if not constant.single:
                journal.time = local_time_to_time(time)
        else:
            if constant.single and time:
                log.warning(f'Ignoring time for {constant.name}')
                time = 0.0
            constant.add_value(s, value, time=time)

    def delete_constant(self, request, s):
        data = request.json
        log.debug(data)
        s.delete(Constant.from_name(s, data))



from logging import getLogger
from os import environ, makedirs
from os.path import exists
from re import sub
from typing import MutableMapping

from .io import clean_path
from .names import BIND, PORT, OFF, LIGHT, DARK, USER, U, PASSWD, P, URI, VERSION, ADMIN_USER, ADMIN_PASSWD

log = getLogger(__name__)


class NamespaceWithVariables(MutableMapping):

    @classmethod
    def _from_ns(cls, ns, env_prefix, version):
        data = vars(ns)
        data[VERSION] = version
        return cls(env_prefix, data)

    def __init__(self, env_prefix, data):
        self.__env_prefix = env_prefix
        self.__dict = {}
        self._set_all(data)

    def _set_all(self, data):
        for name, value in data.items():
            self[name] = value

    def __getitem__(self, name):
        return self.__dict[name]

    def __setitem__(self, name, value):
        value = self.__replace(name, value)
        self.__dict[self.__bar(name)] = value
        self.__dict[self.__unbar(name)] = value

    def __delitem__(self, name):
        del self.__dict[self.__bar(name)]
        del self.__dict[self.__unbar(name)]

    def __iter__(self):
        return iter(self.__dict)

    def __len__(self):
        return len(self.__dict__)

    def _format(self, name=None, value=None):
        if not value: value = self[name]
        while value and '{' in value:
            log.debug(f'Formatting {value}')
            value = value.format(**self)
        return value

    def _format_path(self, name=None, value=None, mkdir=True):
        path = clean_path(self._format(name=name, value=value))
        if not exists(path) and mkdir: makedirs(path, exist_ok=True)
        return path

    def _with(self, **kargs):
        data = dict(self.__dict)
        data.update(kargs)
        return self.__class__(self.__env_prefix, data)

    def __bar(self, name):
        return sub('-', '_', name)

    def __unbar(self, name):
        return sub('_', '-', name)

    def __replace(self, name, value):
        '''Environment overrides command line so that system config overrides user preferences.'''
        env_name = f'{self.__env_prefix}_{self.__bar(name).upper()}'
        if env_name in environ:
            value = environ[env_name]
            log.debug(f'Forcing {name} to {value} via {env_name}')
        return value


def mm(name): return '--' + name


def m(name): return '-' + name


def no(name): return 'no-%s' % name


def add_server_args(cmd, prefix='', default_address='localhost', default_port=80, name=''):
    if prefix: prefix += '-'
    if name: name += ' '
    cmd.add_argument(mm(prefix + BIND), default='localhost', metavar='ADDRESS',
                     help=name + 'bind address' + f' (default {default_address})' if default_address else '')
    cmd.add_argument(mm(prefix + PORT), default=default_port, type=int, metavar='PORT',
                     help=name + f'port' + f' (default {default_port})' if default_port else '')


def add_data_source_args(parser, uri_default):
    parser.add_argument(mm(USER), m(U), default='default', metavar='USER', help='the current user')
    parser.add_argument(mm(PASSWD), m(P), default='', metavar='PASS', help='user password')
    parser.add_argument(mm(ADMIN_USER), default='postgres', metavar='USER', help='the database admin user')
    parser.add_argument(mm(ADMIN_PASSWD), default='', metavar='PASS', help='user database admin password')
    parser.add_argument(mm(URI), default=uri_default, help='use the given database URI')


def color(color):
    if color.lower() not in (LIGHT, DARK, OFF):
        raise Exception(f'Bad color: {color} ({LIGHT}|{DARK}|{OFF})')
    return color



from contextlib import contextmanager
from logging import getLogger
from os import getpid
from sys import argv

from math import floor

log = getLogger(__name__)


def command_root():
    try:
        with open(f'/proc/{getpid()}/cmdline', 'rb') as f:
            line = f.readline()

            def parse():
                word = bytearray()
                for char in line:
                    if char:
                        word.append(char)
                    else:
                        yield word.decode('utf8')
                        word = bytearray()

            words = list(parse())
            log.debug(f'Parsed /proc/{getpid()}/cmdline as {" ".join(words)}')
            if len(argv) > 1:
                i = words.index(argv[1])
                words = words[:i]
            ch2 = ' '.join(words)
            if 'unittest' in ch2:
                log.warning(f'Appear to be inside test runner: {ch2}')
                ch2 = 'python -m ch2'
            log.debug(f'Using command "{ch2}"')
            return ch2
    except:
        log.warning('Cannot read /proc so assuming that ch2 is started on the command line as "ch2"')
        return 'ch2'



from contextlib import contextmanager
from logging import getLogger

from sqlalchemy import create_engine, MetaData, text, inspect
from sqlalchemy.orm import sessionmaker, Session
from uritools import urisplit, uriunsplit

from . import *
from .batch import BatchLoader
from ..commands.args import NO_OP, make_parser, NamespaceWithVariables, PROGNAME, DB_VERSION
from ..common.log import log_current_exception
from ..lib.log import make_log_from_args
from ..lib.utils import grouper

# mention these so they are "created" (todo - is this needed? missing tables seem to get created anyway)

Source,  Interval, Composite, CompositeComponent
ActivityGroup, ActivityJournal, ActivityTimespan, ActivityBookmark
DiaryTopic, DiaryTopicJournal, DiaryTopicField,
ActivityTopic, ActivityTopicJournal, ActivityTopicField,
StatisticName, StatisticJournal, StatisticJournalInteger, StatisticJournalFloat, StatisticJournalText, StatisticMeasure
Pipeline
Constant, SystemConstant, Process
ActivitySimilarity, ActivityNearby
Timestamp, Process, SystemConstant
SectorGroup, Sector, SectorClimb


log = getLogger(__name__)


def scheme(uri):
    return urisplit(uri).scheme


class DirtySession(Session):
    '''Extend Session to record dirty intervals and then mark those intervals when the current transaction ends.'''

    def __init__(self, *args, **kargs):
        super().__init__(*args, **kargs)
        self.__dirty_ids = set()

    def record_dirty_intervals(self, ids):
        self.__dirty_ids.update(ids)

    def __mark_dirty_intervals(self):
        if self.__dirty_ids:
            log.debug(f'Marking {len(self.__dirty_ids)} intervals dirty')
            for ids in grouper(self.__dirty_ids, 900):
                self.query(Interval).filter(Interval.id.in_(ids), Interval.dirty == False). \
                    update({Interval.dirty: True}, synchronize_session=False)
            super().commit()
            self.__dirty_ids = set()

    def commit(self):
        super().commit()
        self.__mark_dirty_intervals()

    def rollback(self):
        super().rollback()
        self.__dirty_ids = set()


class CannotConnect(Exception): pass


class DatabaseBase:

    def __init__(self, uri):
        self.batch = BatchLoader()
        try:
            self.uri = uri
            options = {'echo': False, 'executemany_mode': 'values'}
            connect_args = {}
            uri_parts = urisplit(uri)
            if uri_parts.query:
                for name, value in uri_parts.getquerydict().items():
                    if len(value) > 1: raise Exception(f'Multiple values for option {name}')
                    value = value[0]
                    log.debug(f'Have additional URI option {name} = {value}')
                    if name == 'echo': options[name] = bool(value)
                    elif name == 'executemany_mode': options[name] = value
                    elif name == 'search_path': connect_args['options'] = f'-csearch_path={value}'
                    else: raise Exception(f'Unsupported option {name} = {value}')
                uri = uriunsplit(uri_parts._replace(query=None))
            if uri_parts.scheme != 'postgresql':
                log.warning(f'Legacy scheme {uri_parts.scheme}; discarding options')
                options, connect_args = {}, {}
            log.debug(f'Creating engine for {uri} with options {options} and connect args {connect_args}')
            self.engine = create_engine(uri, **options, connect_args=connect_args)
            self.session = sessionmaker(bind=self.engine, class_=DirtySession)
            self.engine.connect().execute(text('select 1')).fetchone()  # test connection
        except:
            log_current_exception(traceback=False)
            raise CannotConnect(f'Could not connect to database')

    def no_schema(self, table):
        return not inspect(self.engine).has_table(table.__tablename__)

    @contextmanager
    def session_context(self, **kargs):
        session = self.session(**kargs)
        self.batch.enable(session)
        try:
            yield session
            session.commit()
        except:
            session.rollback()
            raise
        finally:
            session.close()

    def __str__(self):
        return f'{self.__class__.__name__} at {self.uri}'


class Database(DatabaseBase):

    def __init__(self, uri):
        super().__init__(uri)

    def no_data(self):
        try:
            with self.session_context() as s:
                n_topics = s.query(DiaryTopicJournal.id).count()
                n_activities = s.query(ActivityJournal.id).count()
                n_statistics = s.query(StatisticJournal.id).count()
                return not (n_topics + n_activities + n_statistics)
        except:
            log_current_exception()
            # log.debug('Discarding error which may contain password')
            return True

    def no_schema(self, table=Constant):
        return super().no_schema(table=table)


def connect_config(args):
    '''
    Bootstrap from commandline-like args.
    '''
    from .config import Config
    if len(args) == 1 and isinstance(args[0], str):
        args = args[0].split()
    elif args:
        args = list(args)
    else:
        args = []
    args.append(NO_OP)
    ns = NamespaceWithVariables._from_ns(make_parser(with_noop=True).parse_args(args), PROGNAME, DB_VERSION)
    make_log_from_args(ns)
    return Config(ns)


def connect(args):
    '''
    Bootstrap from commandline-like args.
    '''
    return connect_config(args).db


class ReflectedDatabase(DatabaseBase):

    def __init__(self, *args, **kargs):
        super().__init__(*args, **kargs)
        self.meta = MetaData()
        self.meta.reflect(bind=self.engine)



from collections import defaultdict
from logging import getLogger

from .args import SOURCE, SEGMENTS, CONSTANTS, KIT, ACTIVITIES, DIARY, SECTORS, infer_flags
from ..common.names import LIST, URI
from ..import_ import available_versions
from ..import_.activity import import_activity
from ..import_.constant import import_constant
from ..import_.diary import import_diary
from ..import_.kit import import_kit
from ..import_.sector import import_sector
from ..lib.log import Record
from ..sql.database import ReflectedDatabase

log = getLogger(__name__)


def import_(config):
    '''
## import

    > ch2 import 0-30

Import data from a previous version (after starting a new version).
Data must be imported before any other changes are made to the database.

By default all types of data (diary, activities, kit, constants and segments) are imported.
Additional flags can enable or disable specific data types.

### Examples

    > ch2 import --diary 0-30

Import only diary entries.

    > ch2 import --disable --diary 0-30

Import everything but diary entries.
    '''
    if config.args[LIST]:
        for uri in available_versions(config):
            print(uri)
    else:
        source = config.args[SOURCE]
        if not source:
            local = list(available_versions(config))
            if local:
                source = local[0]
            else:
                raise Exception('No versions found locally')
        flags = infer_flags(config.args, DIARY, ACTIVITIES, KIT, CONSTANTS, SEGMENTS, SECTORS)
        import_source(config, Record(log), source, flags=flags)


def import_source(config, record, source, flags=None):
    if config.db.no_schema():
        raise Exception(f'Cannot import into a database with no schema (add profile first)')
    # engine needed if source is not a URI
    with record.record_exceptions():
        uri = infer_uri(config, source)
        if flags is None: flags = defaultdict(lambda: True)
        old = ReflectedDatabase(uri)
        if not old.meta.tables:
            record.raise_(f'No tables found in {uri}')
        log.info(f'Importing data from {uri}')
        if flags[DIARY]: import_diary(record, old, config.db)
        if flags[ACTIVITIES]: import_activity(record, old, config.db)
        if flags[KIT]: import_kit(record, old, config.db)
        if flags[CONSTANTS]: import_constant(record, old, config.db)
        if flags[SECTORS]: import_sector(record, old, config.db)


def infer_uri(config, source):
    if ':' in source:
        log.info(f'Using {source} directly as database URI for import')
        return source
    else:
        safe_uri = config.args._with(version=source, passwd='xxxxxx')._format(URI)
        log.info(f'Using {source} as a version number to get URI {safe_uri}')
        return config.args._with(version=source)._format(URI)



from logging import getLogger
from tempfile import TemporaryDirectory

import pandas as pd

from ch2.commands.args import DEV, V, bootstrap_db, BASE, UPLOAD
from ch2.commands.upload import upload
from ch2.common.args import mm, m
from ch2.config.profiles.acooke import acooke
from ch2.data import Statistics
from ch2.names import N
from ch2.pipeline.read.activity import ActivityReader
from tests import LogTestCase, random_test_user

log = getLogger(__name__)


class TestPower(LogTestCase):

    def test_constant(self):

        user = random_test_user()
        with TemporaryDirectory() as f:

            bootstrap_db(user, m(V), '5', mm(DEV), configurator=acooke)

            config = bootstrap_db(user, mm(BASE), f, m(V), '5', mm(DEV), UPLOAD,
                                       'data/test/source/personal/2018-03-04-qdp.fit')
            upload(config)

            with config.db.session_context() as s:
                stats = Statistics(s, activity_journal='2018-03-04 07:16:33', with_timespan=True). \
                    by_name(ActivityReader, N.LATITUDE, N.LONGITUDE, N.SPHERICAL_MERCATOR_X,
                            N.SPHERICAL_MERCATOR_Y, N.DISTANCE, N.ELEVATION, N.SPEED, N.CADENCE, N.HEART_RATE).df
                stats.describe()

                sepn = pd.Series(stats.index).diff().median()  # 7 secs
                start = stats.index.min()  # 2018-03-04 10:16:33+00:00
                finish = stats.index.max()  # 2018-03-04 16:34:51+00:00
                even = pd.DataFrame({'keep': True}, index=pd.date_range(start=start, end=finish, freq=sepn))
                both = stats.join(even, how='outer', sort=True)
                both.interpolate(method='index', limit_area='inside', inplace=True)
                both = both.loc[both['keep'] == True].drop(columns=['keep'])
                both = both.loc[both[N.TIMESPAN_ID].isin(stats[N.TIMESPAN_ID].unique())]
                both.describe()



from logging import getLogger

from sqlalchemy import text

from .db import add_user, add_schema
from .names import URI
from .sql import database_really_exists

log = getLogger(__name__)


class ConnectionError(Exception): pass


def make_user_database(config, user, passwd):

    # this does not create tables, since those are not common to multiple projects

    user_config = config._with(user=user, passwd=passwd)
    user_uri = user_config.args._format(URI)

    if database_really_exists(user_uri):
        log.warning(f'User {user} can already connect')
        return

    cnxn = config.db.engine.connect()
    if cnxn.execute(text('select 1 from pg_roles where rolname = :user'), user=user).fetchone():
        raise ConnectionError(f'User already exists but could not connect with supplied password')

    add_user(user_config)
    add_schema(user_config)
    return user_config



from ch2.common.date import time_to_local_time
from ch2.names import N
from ch2.pipeline.calculate import SummaryCalculator, ElevationCalculator
from ch2.pipeline.calculate.power import PowerCalculator
from ch2.pipeline.read.activity import ActivityReader
from ch2.sql import StatisticName, StatisticJournal, StatisticJournalType
from ch2.sql.types import short_cls


class Statistics:

    def read_all_plottable(self, request, s):
        q = s.query(StatisticName). \
            filter(StatisticName.statistic_journal_type.in_(
            [StatisticJournalType.FLOAT, StatisticJournalType.INTEGER]),
            ~StatisticName.owner.in_([SummaryCalculator, ActivityReader])). \
            order_by(StatisticName.name)
        return [{'name': name.name, 'title': name.title, 'owner': name.owner}
                for name in q.all() if not (
                    (name.owner == short_cls(PowerCalculator)
                     and name.name in [N.POWER_ESTIMATE, N.CLIMB_POWER, N.HEADING, N.VERTICAL_POWER])
                    or
                    (name.owner == short_cls(ElevationCalculator)
                     and name.name in [N.ELEVATION, N.GRADE]))]

    def read_values(self, request, s, name):
        statistic_name = self._resolve_name(request, s, name)
        values = s.query(StatisticJournal). \
            filter(StatisticJournal.statistic_name == statistic_name). \
            order_by(StatisticJournal.time)
        return [{'date': time_to_local_time(value.time),
                 'value': value.value} for value in values]

    def _resolve_name(self, request, s, name):
        owner = request.args.get('owner', None)
        q = s.query(StatisticName).filter(StatisticName.name == name)
        if owner:
            q = q.filter(StatisticName.owner == owner)
        n = q.count()
        if n == 0:
            raise Exception(f'No such statistic ({name}/{owner})')
        if n > 1:
            owners = ', '.join(sn.owner for sn in q.all())
            raise Exception(f'Statistic {name} has multiple owners ({owners})')
        return q.one()




from .achievement import AchievementCalculator
from .activity import ActivityCalculator
from .elevation import ElevationCalculator
from .impulse import ImpulseCalculator
from .kit import KitCalculator
from .nearby import NearbyCalculator
from .response import ResponseCalculator
from .summary import SummaryCalculator
from .sector import SectorCalculator



from logging import getLogger, Formatter, DEBUG, StreamHandler
from logging.handlers import RotatingFileHandler
from sys import exc_info
from traceback import format_tb

from colorlog import ColoredFormatter

from .names import DARK, LIGHT, UNDEF

log = getLogger(__name__)
STDERR_HANDLER = None


def configure_log(name, path, verbosity, levels=None):

    global STDERR_HANDLER
    levels = levels or {}

    if not getLogger(name).handlers:
        file_formatter = Formatter('%(levelname)-7s %(asctime)s: %(message)s')
        file_handler = RotatingFileHandler(path, maxBytes=1e6, backupCount=10)
        file_handler.setLevel(DEBUG)
        file_handler.setFormatter(file_formatter)
        for root, level in levels.items():
            log = getLogger(root)
            log.setLevel(level)
            log.addHandler(file_handler)

        if verbosity:
            stderr_formatter = Formatter('%(levelname)7s: %(message)s')
            STDERR_HANDLER = StreamHandler()
            STDERR_HANDLER.setLevel(10 * (6 - verbosity))
            STDERR_HANDLER.setFormatter(stderr_formatter)
            for root in levels:
                log = getLogger(root)
                log.addHandler(STDERR_HANDLER)


def set_log_color(color):

    if STDERR_HANDLER and color:
        if color.lower() == LIGHT:
            STDERR_HANDLER.setFormatter(
                ColoredFormatter('%(log_color)s%(levelname)8s: %(message_log_color)s%(message)s',
                                 log_colors={'DEBUG': 'black',
                                             'INFO': 'black',
                                             'WARNING': 'black',
                                             'ERROR': 'black',
                                             'CRITICAL': 'black'},
                                 secondary_log_colors={'message':
                                                           {'DEBUG': 'yellow',
                                                            'INFO': 'blue',
                                                            'WARNING': 'black',
                                                            'ERROR': 'bold_red',
                                                            'CRITICAL': 'bold_red'}}))
        elif color.lower() == DARK:
            STDERR_HANDLER.setFormatter(
                ColoredFormatter('%(log_color)s%(levelname)8s: %(message_log_color)s%(message)s',
                                 log_colors={'DEBUG': 'white',
                                             'INFO': 'white',
                                             'WARNING': 'white',
                                             'ERROR': 'white',
                                             'CRITICAL': 'white'},
                                 secondary_log_colors={'message':
                                                           {'DEBUG': 'yellow',
                                                            'INFO': 'white',
                                                            'WARNING': 'cyan',
                                                            'ERROR': 'bold_red',
                                                            'CRITICAL': 'bold_red'}}))


def log_current_exception(traceback=UNDEF, exception_level=DEBUG, traceback_level=DEBUG, first=False):
    from .global_ import global_dev
    if traceback is UNDEF: traceback = global_dev()
    t, e, tb = exc_info()
    try:
        msg = str(e)
        if first: msg = first_line(msg)
        log.log(exception_level, f'Exception: {msg}')
    except:
        pass
    log.log(exception_level, f'Type: {t}')
    if traceback:
        log.log(traceback_level, 'Traceback:\n' + ''.join(format_tb(tb)))


def first_line(exception):
    return str(exception).splitlines()[0]


def log_query(q, label=None):
    if label: log.debug(label)
    log.debug(q.statement.compile(compile_kwargs={"literal_binds": True}))
    return q



from abc import ABC
from collections import defaultdict, namedtuple
from logging import getLogger

from ..common.date import min_time, max_time
from ..common.math import is_nan
from ..sql import StatisticName, Interval, Source
from ..sql.tables.statistic import STATISTIC_JOURNAL_CLASSES

log = getLogger(__name__)


class Loader(ABC):

    def __init__(self, s, owner, add_serial=True, clear_timestamp=True, batch=True):
        self._s = s
        self._owner = owner
        self.__serial = 0 if add_serial else None
        self.__clear_timestamp = clear_timestamp
        self.__batch = batch

        self.__statistic_name_cache = dict()
        self.__source_cache = dict()
        self._staging = defaultdict(list)
        self.__by_name_then_time = defaultdict(dict)
        self.__add_serial = add_serial
        self._start = None
        self._finish = None
        self.__last_time = None
        self.__counts = defaultdict(lambda: 0)

    def load(self):
        if self:
            for type in self._staging:
                log.debug(f'Adding {len(self._staging[type])} instances of {type}')
                for instance in self._staging[type]:
                    self._s.add(instance)
                self._s.commit()
            self._postload()
        else:
            log.warning('No data to load')

    def __bool__(self):
        return bool(self._staging)

    def _postload(self):
        # manually clean out intervals because we're doing a fast load
        if self.__clear_timestamp and self._start and self._finish:
            Interval.record_dirty_times(self._s, self._start, self._finish)
            self._s.commit()

    def add_data(self, name, source, value, time):
        if name in self.__statistic_name_cache:
            statistic_name = self.__statistic_name_cache[name]
        else:
            statistic_name = StatisticName.from_name(self._s, name, self._owner)
            self.__statistic_name_cache[name] = statistic_name

        if is_nan(value):
            raise Exception(f'Bad value for {statistic_name.name}: {value}')

        journal_class = STATISTIC_JOURNAL_CLASSES[statistic_name.statistic_journal_type]

        if self.__add_serial:
            if self.__last_time is None:
                self.__last_time = time
            elif time > self.__last_time:
                self.__last_time = time
                self.__serial += 1
            elif time < self.__last_time:
                raise Exception('Time travel - timestamp for statistic decreased')

        if isinstance(source, Source):
            if source.id not in self.__source_cache:
                self.__source_cache[source.id] = source
        else:
            if source not in self.__source_cache:
                self.__source_cache[source] = Source.from_id(self._s, source)
            source = self.__source_cache[source]

        self._start = min_time(self._start, time)
        self._finish = max_time(self._finish, time)

        # set statistic_name and source (as well as ids) so that we can correctly test in
        # Source for dirty intervals
        instance = journal_class(statistic_name=statistic_name, statistic_name_id=statistic_name.id,
                                 source=source, source_id=source.id, value=value, time=time, serial=self.__serial)

        if instance.time in self.__by_name_then_time[statistic_name.name]:
            previous = self.__by_name_then_time[statistic_name.name][instance.time]
            if instance.value == previous.value:
                log.warning(f'Discarding duplicate for {statistic_name.name} at {instance.time} '
                            f'(value {instance.value})')
            else:
                self._resolve_duplicate(statistic_name.name, instance, previous)
            return
        else:
            self.__by_name_then_time[statistic_name.name][instance.time] = instance

        self._staging[journal_class].append(instance)
        self.__counts[statistic_name.name] += 1

    def _resolve_duplicate(self, name, instance, prev):
        raise Exception(f'Conflict at ({instance.time}) for {name} '
                        f'(values {instance.value}/{prev.value})')

    def as_waypoints(self, names):
        Waypoint = make_waypoint(names.values())
        time_to_waypoint = defaultdict(lambda: Waypoint())
        for type in self._staging:
            for sjournal in self._staging[type]:
                name = sjournal.statistic_name.name
                if name in names:
                    time_to_waypoint[sjournal.time] = \
                        time_to_waypoint[sjournal.time]._replace(**{'time': sjournal.time,
                                                                    names[name]: sjournal.value})
        return [time_to_waypoint[time] for time in sorted(time_to_waypoint.keys())]

    def coverage_percentages(self):
        total = max(self.__counts.values())
        for name, count in self.__counts.items():
            yield name, 100 * count / total


def make_waypoint(names, extra=None):
    names = list(names)
    if extra:
        names += [extra]
    names = ['time'] + names
    defaults = [None] * len(names)
    return namedtuple('Waypoint', names, defaults=defaults)



from logging import getLogger

import pandas as pd
from geoalchemy2 import WKBElement
from geoalchemy2.shape import to_shape
from sqlalchemy import text

from . import ContentType
from ...names import N
from ...pipeline.calculate.elevation import expand_distance_time
from ...sql.tables.sector import SectorJournal
from ...sql.utils import WGS84_SRID

log = getLogger(__name__)


class Route(ContentType):

    def read_activity_latlon(self, request, s, activity):
        df = expand_distance_time(self._read_activity_route(s, activity))
        latlon = list(df[[N.LATITUDE, N.LONGITUDE]].itertuples(index=None, name=None))
        elevation = list(df[[N.DISTANCE, N.ELEVATION]].itertuples(index=None, name=None))
        return {'latlon': latlon,
                'elevation': elevation,
                'sectors': list(self._read_sectors(s, activity))}

    def _read_activity_route(self, s, activity_journal_id):
        sql = text(f'''
  with points as (select st_dumppoints(aj.route_edt::geometry) as point
                    from activity_journal as aj
                   where aj.id = :activity_journal_id)
select st_x((point).geom) as {N.LONGITUDE}, st_y((point).geom) as {N.LATITUDE}, 
       st_z((point).geom) as {N.ELEVATION}, st_m((point).geom) as "{N.DISTANCE_TIME}"
  from points;
        ''')
        log.debug(sql)
        return pd.read_sql(sql, s.connection(),
                           params={'activity_journal_id': activity_journal_id})

    def _read_sectors(self, s, activity):
        for sjournal in s.query(SectorJournal).filter(SectorJournal.activity_journal_id == activity).all():
            yield {'latlon': self._read_sector_route(s, sjournal.sector.id),
                   'type': sjournal.sector.type,
                   'db': sjournal.sector.id}  # todo - probably want tuple w 'climb' etc here

    def _read_sector_route(self, s, sector_id):
        q = text(f'''
select st_transform(st_setsrid(s.route, sg.srid), {WGS84_SRID})
  from sector as s,
       sector_group as sg
 where s.id = :sector_id''')
        return self._wkb_to_latlon(WKBElement(s.connection().execute(q, sector_id=sector_id).fetchone()[0]))

    def _wkb_to_latlon(self, wkb):
        return [(lat, lon) for (lon, lat) in to_shape(wkb).coords]

    def read_sector_latlon(self, request, s, sector):
        return {'latlon': self._read_sector_route(s, sector)}




from logging import getLogger

from .args import LIKE, WORKER, ARG, parse_pairs, KARG, FORCE, CPROFILE
from ..common.args import mm
from ..pipeline.process import run_pipeline
from ..sql.tables.pipeline import PipelineType

log = getLogger(__name__)


def process(config):
    '''
## process

    > ch2 process

Update the database, reading news files and calculating missing statistics.

    > ch2 process --force

Delete statistics and then calculate new values.

    > ch2 --dev calculate --like '%Activity%' --force 2020-01-01 -Kn_cpu=1

Calculate activity statistics from 2020 onwards in a single process for debugging.
    '''
    args = config.args
    if bool(args[WORKER]) != bool(args[ARG]):
        raise Exception(f'{mm(WORKER)} and arguments should be used together')
    if args[LIKE] and args[WORKER]:
        raise Exception(f'{mm(LIKE)} cannot be used with {mm(WORKER)}')
    run_pipeline(config, PipelineType.PROCESS, *args[ARG],
                 like=args[LIKE], worker=args[WORKER], cprofile=args[CPROFILE],
                 **parse_pairs(args[KARG]))



import datetime as dt
from json import loads
from logging import getLogger

import pandas as pd
from sqlalchemy import text, func

from .elevation import expand_distance_time, elapsed_time_to_time
from .utils import ActivityGroupProcessCalculator
from ...common.date import local_time_to_time
from ...common.log import log_current_exception
from ...data.sector import add_start_finish
from ...names import N
from ...sql import Timestamp, Constant, Sector, SectorGroup, SectorClimb
from ...sql.types import linestringxy
from ...sql.utils import add

log = getLogger(__name__)

DISTINCT_CLIMB = 0.5


class FindClimbCalculator(ActivityGroupProcessCalculator):

    def __init__(self, *args, climb=None, **kargs):
        super().__init__(*args, **kargs)
        self.__climb_ref = climb

    def _startup(self, s):
        from ...data.climb import Climb
        super()._startup(s)
        self.__climb = Climb(**loads(Constant.from_name(s, self.__climb_ref).at(s).value))

    def _shutdown(self, s):
        super()._shutdown(s)
        if not self.worker:
            log.info('Checking for extra prunes')
            n_climbs, n_prunes = 0, 0
            for row in s.query(SectorClimb.id).all():
                if self.__prune_climb(s, row.id):
                    n_prunes += 1
                else:
                    n_climbs += 1
            log.info(f'Total {n_climbs} climbs ({n_prunes} pruned)')

    def _run_one(self, missed):
        start = local_time_to_time(missed)
        with self._config.db.session_context() as s:
            ajournal = self._get_source(s, start)
            with Timestamp(owner=self.owner_out, source=ajournal).on_success(s):
                if ajournal.route_edt:
                    for sector_group in s.query(SectorGroup). \
                            filter(func.st_distance(SectorGroup.centre, ajournal.centre)
                                   < SectorGroup.radius). \
                            all():
                        log.info(f'Finding climbs for activity journal {ajournal.id} / sector group {sector_group.id}')
                        try:
                            self.__find_big_climbs(s, sector_group, ajournal)
                        except Exception as e:
                            log.warning(f'Climb detection failed with {e} for activity journal {ajournal.id} '
                                        f'and sector group {sector_group.id}')
                            log_current_exception(True)
                            s.rollback()

    def __find_big_climbs(self, s, sector_group, ajournal):
        from ...data.climb import find_climbs
        df = self.__read_complete_route(s, sector_group.id, ajournal.id)
        df = expand_distance_time(df)
        df = elapsed_time_to_time(df, ajournal.start)
        df = df.set_index(df[N.TIME]).drop(columns=[N.TIME])
        for climb in find_climbs(df, params=self.__climb):
            log.info(f'Have a climb (elevation {climb[N.CLIMB_ELEVATION]})')
            self.__register_climb(s, df, climb, sector_group, ajournal.id)

    def __read_complete_route(self, s, sector_group_id, activity_journal_id):
        sql = text(f'''
  with points as (select st_dumppoints(st_transform(aj.route_edt::geometry, sg.srid)) as point
                    from activity_journal as aj,
                         sector_group as sg
                   where aj.id = :activity_journal_id
                     and sg.id = :sector_group_id
                     and st_distance(sg.centre, aj.centre) < sg.radius)
select st_x((point).geom) as x, st_y((point).geom) as y, 
       st_z((point).geom) as {N.ELEVATION}, st_m((point).geom) as "{N.DISTANCE_TIME}"
  from points;
        ''')
        log.debug(sql)
        df = pd.read_sql(sql, s.connection(),
                         params={'sector_group_id': sector_group_id,
                                 'activity_journal_id': activity_journal_id})
        return df

    def __register_climb(self, s, df, climb_id, sector_group, activity_journal_id):
        # this is not easy
        # we cannot exclude climbs based on a constraint (it's too complex)
        # we are loading in parallel, so don't have a clear idea of what other climbs exist at any one point in time
        # so what we do is load all climbs and then delete conflicts
        # deletion always respects earlier (smaller ID) climbs, so we get something similar(?) to what we
        # would get if all loaded in order
        df = df.loc[climb_id[N.TIME] - dt.timedelta(seconds=climb_id[N.CLIMB_TIME]): climb_id[N.TIME]]
        route = linestringxy([(row.x, row.y) for row in df.itertuples()], type='geometry')
        climb_id = self.__add_climb(s, sector_group, climb_id, route, activity_journal_id)
        add_start_finish(s, climb_id)
        log.debug(f'Added climb {climb_id}')
        self.__prune_climb(s, climb_id)
        s.commit()

    def __add_climb(self, s, sector_group, climb, route, activity_journal_id):
        if N.CLIMB_CATEGORY in climb:
            title = f'Climb (cat {climb[N.CLIMB_CATEGORY]})'
        else:
            title = f'Climb (uncat)'
        category = climb.get(N.CLIMB_CATEGORY, None)
        # text because we're passing in direct SQL functions, not EWKT
        climb = add(s, SectorClimb(sector_group=sector_group, route=text(route), title=title, owner=self,
                                   distance=climb[N.CLIMB_DISTANCE], category=category,
                                   elevation=climb[N.CLIMB_ELEVATION]))
        s.flush()
        return climb.id

    def __prune_climb(self, s, climb_id):
        sql = text(f'''
  with candidates as (select st_length(st_setsrid(b.route, sg.srid)) as l1,
                             st_length(st_difference(st_setsrid(b.route, sg.srid), st_setsrid(s.hull, sg.srid))) as l2,
                             st_length(st_setsrid(s.route, sg.srid)) as l3,
                             st_length(st_difference(st_setsrid(s.route, sg.srid), st_setsrid(b.hull, sg.srid))) as l4
                        from sector as b,
                             sector as s,
                             sector_group as sg
                       where s.id = :climb_id
                         and b.id < s.id
                         and sg.id = s.sector_group_id
                         and sg.id = b.sector_group_id
                         and st_intersects(st_setsrid(b.route, sg.srid), st_setsrid(s.hull, sg.srid)))
select count(1)
  from candidates as c
 where l2 / l1 < {DISTINCT_CLIMB}
   and l4 / l3 < {DISTINCT_CLIMB}
''')
        if s.connection().execute(sql, climb_id=climb_id).scalar():
            log.debug(f'Deleting climb {climb_id}')
            s.query(Sector).filter(Sector.id == climb_id).delete(synchronize_session=False)
            return True
        else:
            return False



import numpy as np
from bokeh.io import show, output_file
from bokeh.layouts import gridplot
from scipy.integrate import trapz
from scipy.interpolate import interp1d

from ch2.data import *
from ch2.data.impulse import hr_zone, impulse_10
from ch2.data.power import add_differentials, add_energy_budget, add_loss_estimate, add_power_estimate
from ch2.jupyter.decorator import template
from ch2.names import N
from ch2.pipeline.calculate.impulse import HRImpulse
from ch2.pipeline.owners import *
from ch2.sql import ActivityJournal


@template
def power_v_hr(local_time, activity_group):

    f'''
    # Power v HR ({local_time}, {activity_group})
    '''

    '''
    $contents
    '''

    '''
    ## Load Data
    
    Open a connection to the database and load the data we require.
    '''

    local_times = ['2021-04-18 05:50:52', '2021-04-11 05:52:16', '2021-04-03 05:59:03']
    dt = 60

    s = session('-v2')
    # stats = Statistics(s, activity_journal=local_time, activity_group=activity_group, with_timespan=True)
    sources = [ActivityJournal.at(s, local_time, activity_group=activity_group) for local_time in local_times]
    stats = Statistics(s, sources=sources, with_timespan=True, with_sources=True)
    df = stats.by_name(ElevationCalculator, N.ELEVATION, N.GRADE). \
        by_name(ActivityReader, N.DISTANCE, N.SPEED, N.HEART_RATE). \
        with_.transform(N.DISTANCE, scale=1000). \
        concat(N.DISTANCE, 100). \
        concat_index(300). \
        without_sources().df
    # TODO - change default to use just the N.HR_IMPULSE_10
    fthr_df = Statistics(s).by_name(Constant, N.FTHR).df
    hr_zones = hr_zones_from_database(s, local_time, activity_group)

    hr_zone(df, fthr_df)
    ldf = linear_resample_time(df, dt=dt)
    ldf = add_differentials(ldf, max_gap=1.1 * dt)

    '''
    ## Define Model
    '''

    mass = 65 + 9
    params = {'cda': 0.42, 'crr': 0.0055, 'gamma': 2, 'zero': 2}
    bounds = {'cda': (0.1, 0.5), 'crr': (0.0, 0.01), 'gamma': (0.5, 2.5), 'zero': (1, 3)}

    def model(params, df):
        idf = impulse_10(df, HRImpulse('HR Impulse', gamma=params['gamma'], zero=params['zero'], one=6, max_secs=60))
        idf.rolling('60S').apply(lambda g: trapz(g, g.index.view(np.int64) / 10**9))
        mdf = ldf.copy(deep=True)
        mdf[N.HR_IMPULSE_10] = interp1d(idf.index.view(np.int64), idf[N.HR_IMPULSE_10], bounds_error=False)(mdf.index.view(np.int64))
        mdf = add_energy_budget(mdf, mass)
        mdf = add_loss_estimate(mdf, mass, cda=params['cda'], crr=params['crr'])
        mdf = add_power_estimate(mdf)
        mdf[N.POWER_ESTIMATE].clip(lower=0, inplace=True)
        return mdf

    def cost(mdf):
        # this fits a line through the origin (the param 'zero' is the offset, effectively)
        mdf.dropna(inplace=True)
        x = np.array(mdf[N.HR_IMPULSE_10])[:, np.newaxis]
        y = mdf[N.POWER_ESTIMATE]
        _, r, _, _ = np.linalg.lstsq(x, y, rcond=None)
        return r[0]

    '''
    ## Check
    '''

    output_file(filename='/dev/null')

    def plot(mdf):

        el = comparison_line_plot(700, 200, N.DISTANCE, N.ELEVATION, mdf)
        xrange = el.x_range if el else None

        hri = comparison_line_plot(700, 200, N.DISTANCE, N.HR_IMPULSE_10, mdf, ylo=0, x_range=xrange)
        xrange = xrange or (hri.x_range if hri else None)

        hr = comparison_line_plot(700, 200, N.DISTANCE, N.HEART_RATE, mdf, x_range=xrange)
        add_hr_zones(hr, mdf, N.DISTANCE, hr_zones)
        xrange = xrange or (hr.x_range if hr else None)

        pw = comparison_line_plot(700, 200, N.DISTANCE, N.POWER_ESTIMATE, mdf, ylo=0, x_range=xrange)
        pw.varea(source=mdf, x=N.DISTANCE, y1=0, y2=N.VERTICAL_POWER,
                 level='underlay', color='black', fill_alpha=0.25)

        show(gridplot([[el], [hri], [hr], [pw]]))

    plot(model(params, df))

    '''
    ## Fit Data
    '''

    params = {'cda': 0.42, 'crr': 0.0055, 'gamma': 2, 'zero': 2}
    fit(params, df, model, cost, bounds=bounds)
    plot(model(params, df))
    params



from logging import getLogger

log = getLogger(__name__)

__GLOBAL_DEV = False


def set_global_dev(dev):
    global __GLOBAL_DEV
    __GLOBAL_DEV = dev
    log.debug(f'Setting global dev flag: {dev}')


def global_dev():
    global __GLOBAL_DEV
    return __GLOBAL_DEV



from logging import getLogger

from werkzeug import Response
from werkzeug.wrappers import ETagResponseMixin

from . import ContentType
from ...commands.args import IMAGE_DIR

log = getLogger(__name__)


class CacheResponse(Response, ETagResponseMixin):
    pass


class BaseImage(ContentType):

    def __init__(self, config):
        self._image_dir = config.args._format_path(IMAGE_DIR)

    def _serve(self, path):
        try:
            log.debug(f'Reading {path}')
            with open(path, 'rb') as input:
                response = CacheResponse(input.read())
            self.set_content_type(response, path)
            response.cache_control.max_age = 3600
            return response
        except Exception as e:
            log.warning(f'Error serving {path}: {e}')
            raise


class Thumbnail(BaseImage):

    def __call__(self, request, s, activity, sector=None):
        from ...commands.thumbnail import create_in_cache
        path = create_in_cache(self._image_dir, s, activity, sector_id=sector)
        return self._serve(path)


class Sparkline(BaseImage):

    def __call__(self, request, s, statistic, sector, activity, invert=False):
        from ...commands.sparkline import create_in_cache
        path = create_in_cache(self._image_dir, s, statistic, sector_id=sector, activity_id=activity, invert=invert)
        return self._serve(path)



from ch2.commands.args import V, DEV, bootstrap_db
from ch2.common.args import mm, m
from ch2.common.date import to_date, add_date
from ch2.config.plan.british import twelve_week_improver
from ch2.config.plan.exponential import exponential_time, exponential_distance
from ch2.config.profiles.default import default
from ch2.sql.tables.topic import DiaryTopic
from tests import LogTestCase, random_test_user


class TestPlan(LogTestCase):

    def test_british(self):
        user = random_test_user()
        config = bootstrap_db(user, m(V), '5', mm(DEV), configurator=default)
        plan = twelve_week_improver('2018-07-25')
        plan.create(config.db)
        with config.db.session_context() as s:
            root = s.query(DiaryTopic).filter(DiaryTopic.parent_id == None, DiaryTopic.title == 'Plan').one()
            self.assertEqual(len(root.children), 1)
            self.assertTrue(root.schedule)
            self.assertEqual(root.schedule.start, to_date('2018-07-25'))
            self.assertEqual(root.schedule.finish, add_date('2018-07-25', (12, 'w')))
            parent = root.children[0]
            self.assertEqual(len(parent.children), 7)
            for child in parent.children:
                print(child)

    def test_exponential_time(self):
        user = random_test_user()
        config = bootstrap_db(user, m(V), '5', mm(DEV), configurator=default)
        plan = exponential_time('Time test', '2d[2]', '20M', 5, '2018-07-25', '3m')
        plan.create(config.db)
        with config.db.session_context() as s:
            root = s.query(DiaryTopic).filter(DiaryTopic.parent_id == None, DiaryTopic.title == 'Plan').one()
            self.assertEqual(len(root.children), 1)
            parent = root.children[0]
            self.assertEqual(len(parent.children), 46)
            for child in parent.children:
                print(child)

    def test_exponential_distance(self):
        user = random_test_user()
        config = bootstrap_db(user, m(V), '5', mm(DEV), configurator=default)
        plan = exponential_distance('Distance test', 'w[mon,wed,fri]', '20km', 5, '2018-07-25', '1m')
        plan.create(config.db)
        with config.db.session_context() as s:
            root = s.query(DiaryTopic).filter(DiaryTopic.parent_id == None, DiaryTopic.title == 'Plan').one()
            self.assertEqual(len(root.children), 1)
            parent = root.children[0]
            self.assertEqual(len(parent.children), 14)
            for child in parent.children:
                print(child)



from glob import glob
from logging import getLogger
from os import makedirs
from os.path import basename, join, exists, dirname

from ..commands.args import KIT, PATH, DATA, UPLOAD, PROCESS, CPROFILE
from ..common.date import time_to_local_time, Y, YMDTHMS, to_time
from ..common.io import touch, clean_path, data_hash
from ..common.log import log_current_exception
from ..fit.format.read import filtered_records
from ..fit.format.records import merge_duplicates, fix_degrees, unpack_single_bytes
from ..lib.io import split_fit_path
from ..lib.log import Record
from ..lib.utils import timing
from ..pipeline.process import run_pipeline
from ..pipeline.read.utils import AbortImportButMarkScanned, ProcessFitReader
from ..sql import KitItem, FileHash, PipelineType

log = getLogger(__name__)


TYPE = 'type'
TIME = 'time'
STREAM = 'stream'
DATA = 'data'
EXTRA = 'extra'
HASH = 'hash'
NAME = 'name'
ACTIVITY = 'activity'
MONITOR = 'monitor'
SPORT = 'sport'
DIR = 'dir'
DOT_FIT = '.fit'
READ_PATH = 'read-path'
WRITE_PATH = 'write-path'

MONITORING_INFO_ATTR = 'monitoring_info'


def upload(config):
    '''
## upload

    > ch2 upload --kit ITEM [ITEM...] -- PATH [PATH ...]

Copy FIT files, storing the data in the permanent store on the file system.
Optionally, call process after, to add data to the database.
Both monitor and activity files are accepted.

### Examples

    > ch2 upload --kit cotic -- ~/fit/2018-01-01.fit

will store the given file, add activity data to the database (associated with the kit 'cotic'), check for
new monitor data, and update statistics.
    '''
    args = config.args
    with timing(UPLOAD):
        upload_files(Record(log), config, files=open_files(args[PATH]), items=args[KIT])
        if args[PROCESS]:
            run_pipeline(config, PipelineType.PROCESS, cprofile=args[CPROFILE])


class SkipFile(Exception):
    pass


def open_files(paths):
    # use an iterator here to avoid opening too many files at once
    for path in paths:
        path = clean_path(path)
        if exists(path):
            name = basename(path)
            log.debug(f'Reading {path}')
            stream = open(path, 'rb')
            yield {STREAM: stream, NAME: name, READ_PATH: path}
        else:
            raise Exception(f'No file at {path}')


def check_items(s, items):
    for item in items:
        if not s.query(KitItem).filter(KitItem.name == item).one_or_none():
            raise Exception(f'Kit item {item} does not exist')


def read_file(file):
    # add DATA to dicts with STREAM
    log.debug(f'Reading {file[NAME]}')
    file[DATA] = file[STREAM].read()
    file[STREAM].close()


def hash_file(file):
    # add HASH to dicts with DATA
    log.debug(f'Hashing {file[NAME]}')
    file[HASH] = data_hash(file[DATA])
    log.debug(f'Hash of {file[NAME]} is {file[HASH]}')


def check_path(file):
    path, name = file[WRITE_PATH], file[NAME]
    gpath, _ = split_fit_path(path)
    match = glob(gpath)
    if match:
        touch(match[0])  # trigger re-processing
        raise SkipFile(f'A file already exists for {name} at {match[0]}')
    log.debug(f'File {name} cleared for path {path}')


def check_hash(s, file):
    hash, path, name = file[HASH], file[WRITE_PATH], file[NAME]
    file_hash = s.query(FileHash).filter(FileHash.hash == hash).one_or_none()
    if file_hash and file_hash.file_scan:
        raise SkipFile(f'A file was already processed with the same hash as {name} '
                       f'({hash} at {file_hash.file_scan.path})')
    log.debug(f'File {name} cleared for hash {hash}')


def check_file(s, file):
    check_path(file)
    check_hash(s, file)


def _parse_records(data):
    types, messages, records = filtered_records(data)
    return [record.as_dict(merge_duplicates, fix_degrees, unpack_single_bytes)
            for _, _, record in sorted(records,
                                       key=lambda r: r[2].timestamp if r[2].timestamp else to_time(0.0))]


def _read_first_timestamp(path, records):
    return ProcessFitReader.assert_contained(path, records, [MONITORING_INFO_ATTR], 0)


def parse_fit_data(file, items=None):
    from ch2.pipeline.read.activity import ActivityReader
    try:
        # add TIME and TYPE and EXTRA (and maybe SPORT) given (fit) DATA and NAME
        try:
            records = _parse_records(file[DATA])
            file[TIME] = _read_first_timestamp(file[NAME], records)
            file[TYPE] = MONITOR
            # don't use '-' here or it will be treated as kit in path matching
            file[EXTRA] = ':' + file[HASH][0:5]
            log.debug(f'File {file[NAME]} contains monitor data')
        except AbortImportButMarkScanned:
            records = ActivityReader.parse_records(file[DATA])
            file[TIME] = ActivityReader.read_first_timestamp(file[NAME], records)
            file[SPORT] = ActivityReader.read_sport(file[NAME], records)
            file[TYPE] = ACTIVITY
            file[EXTRA] = '-' + ','.join(items) if items else ''
            log.debug(f'File {file[NAME]} contains activity data')
    except Exception as e:
        log_current_exception()
        raise Exception(f'Could not parse {file[NAME]} as a fit file')


def build_path(data_dir, file):
    # add PATH given TIME, TYPE, SPORT and EXTRA
    date = time_to_local_time(file[TIME], fmt=YMDTHMS)
    year = time_to_local_time(file[TIME], fmt=Y)
    file[DIR] = join(data_dir, file[TYPE], year)
    if SPORT in file: file[DIR] = join(file[DIR], file[SPORT])
    file[WRITE_PATH] = join(file[DIR], date + file[EXTRA] + DOT_FIT)
    log.debug(f'Target directory is {file[DIR]}')


def write_file(file):
    try:
        dir = dirname(file[WRITE_PATH])
        if not exists(dir):
            log.debug(f'Creating {dir}')
            makedirs(dir)
        if exists(file[WRITE_PATH]):
            log.warning(f'Overwriting data at {file[WRITE_PATH]}')
        with open(file[WRITE_PATH], 'wb') as out:
            log.info(f'Writing {file[NAME]} to {file[WRITE_PATH]}')
            out.write(file[DATA])
    except Exception as e:
        log_current_exception()
        raise Exception(f'Could not save {file[NAME]}')


def upload_files(record, data, files=tuple(), items=tuple()):
    with record.record_exceptions():
        with data.db.session_context() as s:
            data_dir = data.args._format_path(DATA)
            check_items(s, items)
            for file in files:
                try:
                    read_file(file)
                    hash_file(file)
                    parse_fit_data(file, items=items)
                    build_path(data_dir, file)
                    check_file(s, file)
                    write_file(file)
                    record.info(f'Uploaded {file[NAME]} to {file[WRITE_PATH]}')
                except SkipFile as e:
                    record.warning(e)



from collections import namedtuple
from logging import getLogger

from geoalchemy2.shape import to_shape
from math import log10
from sqlalchemy import desc, text

from ..data.sector import add_start_finish
from ..sql import ActivityJournal, ClusterInputScratch, ClusterHull, ClusterFragmentScratch, SectorGroup, Sector
from ..sql.tables.sector import SectorType
from ..sql.types import short_cls

log = getLogger(__name__)


'''
what do we want to do here?
somehow we want to automatically identify segments of a ride.
what is a segment?
one answer is to construct a graph of rides where intersections between paths are nodes.
then a segment is an edge between two nodes (or possibly a popular series of edges).
how do we construct that graph?  
or even something vaguely similar?
these notes are written in retrospect (in a pause during development).
this work was not done in response to a simple analysis - it was an exploration of what is possible.
if we treat all the gps points as a cloud of data then we can 
1 - find the densest regions by doing cluster analysis on a sparse sample of data
2 - iteratively find less dense regions by sampling more frequently
3 - once we are sampling all points, we can relax conditions to fill in what is left
this gives us a hierarchy of clusters.
is that enough to construct the graph?
i don't know - i haven't got that far yet.
the above process is slow and noisy - how can we improve it?
one idea is to use small line segments rather than points,
this requires surprisingly little modification of the code. 
it runs faster and produces fewer clusters (with similar total coverage) but is not so great at infrequently used routes.
maybe the infrequent routes don't matter so much?
in general, the larger the segment, the faster the time but the lower the resolution.
having an overlap seems to help 'semantic' groups.
once we have local clusters, find 'typical' routes and then group by those?
'''


def hulls_from_last_activity(s, radius_km):
    activity_journal = s.query(ActivityJournal). \
        filter(ActivityJournal.centre != None). \
        order_by(desc(ActivityJournal.start)).first()
    if activity_journal:
        return hulls_from_activity(s, activity_journal, radius_km)
    else:
        log.warning('No activities, so no clusters (no sector group)')


def hulls_from_activity(s, activity_journal, radius_km):
    return hulls_from_point(s, activity_journal.centre, radius_km,
                            title=f'From activity {activity_journal.start}')


Parameters = namedtuple('Parameters', 'start,exp,min_dbscan,min_total,target,buffer,indep,overlap',
                        defaults=(1000, 2.0, 2, 100, 0.75, 10, 4, 0))


def hulls_from_point(s, centre, radius_km, title, parameters=Parameters()):
    sector_group = SectorGroup.add(s, to_shape(centre).coords[0], radius_km, title)
    log.info(f'Starting clustering for {sector_group.title}')
    delete_tmp_lines(s, sector_group)
    populate_tmp_lines(s, sector_group, parameters)
    delete_hulls(s, sector_group)
    eps = 20
    for log_nth in range(int(0.5 + log10(parameters.start) / log10(parameters.exp)), -1, -1):
        nth = int(0.5 + parameters.exp ** log_nth)
        cluster_remaining(s, sector_group, parameters, nth, eps)
    for nth in range(2, 6):
        eps = 10 * 2 ** nth
        cluster_remaining(s, sector_group, parameters, -1 * nth, eps)
    log.info(f'Finished clustering for {sector_group.title}')
    delete_tmp_lines(s, sector_group)
    return sector_group


def delete_hulls(s, sector_group):
    query = s.query(ClusterHull).filter(ClusterHull.sector_group_id == sector_group.id)
    n = query.count()
    if n:
        log.warning(f'Deleting {n} hulls for {sector_group.title}')
        query.delete(synchronize_session=False)


def cluster_remaining(s, sector_group, parameters, nth, eps):
    log.info(f'Grouping at {nth}/{eps}')
    identify_groups(s, sector_group, parameters, nth, eps)
    label_all_geoms(s, sector_group, parameters, nth, eps, max(nth, 1) * parameters.min_total)


def identify_groups(s, sector_group, parameters, nth, eps):
    sql = text('''
  with groups as (select id,
                         st_clusterdbscan(geom, eps:=:eps, minpoints:=:min_dbscan) over () as "group"
                    from (select id,
                                 row_number() over () as i,
                                 geom
                            from cluster_input_scratch
                           where "group" is null
                             and sector_group_id = :sector_group_id
                           order by id) as x
                   where i % greatest(:nth, 1) = 0)
update cluster_input_scratch
   set "group" = groups."group",
       level = :nth
  from groups
 where cluster_input_scratch.id = groups.id
   and groups."group" is not null;
''')
    log.debug(sql)
    s.connection().execute(sql, sector_group_id=sector_group.id, nth=nth, eps=eps,
                           min_dbscan=parameters.min_dbscan)
    s.commit()


def label_all_geoms(s, sector_group, parameters, nth, eps, min_total):
    sql = text('''
  with points as (select st_dumppoints(geom) as point,
                         "group"
                    from cluster_input_scratch
                   where level = :nth
                     and sector_group_id = :sector_group_id),
       hulls as (select st_buffer(st_concavehull(st_collect((point).geom), :target), :buffer) as hull,
                        "group"
                   from points
                  group by "group"),
       census as (select count(c.geom) as n,
                         h."group",
                         h.hull
                    from cluster_input_scratch as c,
                         hulls as h
                   where st_covers(h.hull, c.geom)
                     and (c."group" = h."group" or c."group" is null)
                   group by h.hull, h."group"
                   order by count(c.geom) desc)
insert into cluster_hull ("group", sector_group_id, level, hull)
select cs."group", :sector_group_id, :nth, cs.hull
  from census as cs
 where cs.n >= :min_total;
 
update cluster_input_scratch as c
   set "group" = h."group",
       level = h.level
  from cluster_hull as h
 where st_covers(h.hull, c.geom)
   and (c.level is null or c.level = h.level)
   and c.sector_group_id = h.sector_group_id
   and h.level = :nth
   and h.sector_group_id = :sector_group_id;
   
update cluster_input_scratch as c 
   set "group" = null,
       level = null
 where c.level = :nth
   and c.sector_group_id = :sector_group_id
   and c."group" not in (
       select "group" from cluster_hull where sector_group_id = :sector_group_id and level = :nth)
    ''')
    log.debug(sql)
    s.connection().execute(sql, sector_group_id=sector_group.id, nth=nth, eps=eps,
                           min_dbscan=parameters.min_dbscan, min_total=min_total,
                           target=parameters.target, buffer=parameters.buffer)
    s.commit()


def populate_tmp_lines(s, sector_group, parameters):
    '''
    segments of length indep+overlap, where indep points are independent and overlap points overlap.
    indep=1 overlap=0 - each point is separate  o o o o
    indep=2 overlap=0 - pairs  o-o o-o o-o
    indep=2 overlap=1 - overlapping triplets o-o-O-o-o-O-o-o-O
    '''
    sql = text('''
  with lines as (select aj.id, 
                        st_force3dm(st_transform(aj.route_et::geometry, sg.srid)) as line
                   from activity_journal as aj,
                        sector_group as sg
                  where sg.id = :sector_group_id
                    and st_distance(sg.centre, aj.centre) < sg.radius),
       lengths as (select id, st_npoints(line) - :indep - :overlap as length from lines),
       series as (select id, generate_series(1, length, :indep) as series from lengths),
       pairs as (select st_makeline(st_pointn(line, series), st_pointn(line, series + :indep + :overlap)) as pair
                   from series inner join lines on lines.id = series.id)
insert into cluster_input_scratch (geom, sector_group_id)
select pair, :sector_group_id
  from pairs;  
''')
    log.debug(sql)
    s.connection().execute(sql, sector_group_id=sector_group.id, indep=parameters.indep, overlap=parameters.overlap)
    s.commit()


def delete_tmp_lines(s, sector_group):
    s.query(ClusterInputScratch). \
        filter(ClusterInputScratch.sector_group_id == sector_group.id). \
        delete(synchronize_session=False)


def sectors_from_hulls(s, sector_group):
    from ..pipeline.calculate.cluster import ClusterCalculator
    log.info(f'Finding sectors for clusters for {sector_group.title}')
    delete_tmp_fragments(s, sector_group)
    populate_tmp_fragments_from_hulls(s, sector_group)
    delete_sectors(s, sector_group)
    identify_sectors(s, sector_group)
    for id in s.query(Sector.id). \
            filter(Sector.sector_group_id == sector_group.id,
                   Sector.start == None,
                   Sector.finish == None,
                   Sector.owner == short_cls(ClusterCalculator)).all():
        add_start_finish(s, id[0])
    delete_tmp_fragments(s, sector_group)
    log.info(f'Finished finding sectors for clusters for {sector_group.title}')


def populate_tmp_fragments_from_hulls(s, sector_group, min_separation=50):
    sql = text('''
  with route as (select st_transform(route_a::geometry, sg.srid) as route,
                        aj.id as activity_journal_id
                   from activity_journal as aj,
                        sector_group as sg
                  where sg.id = :sector_group_id
                    and st_distance(aj.centre, sg.centre) < sg.radius),
       straight as (select (st_dump(st_multi(st_locatebetween(r.route, -2, 2)))).geom as straight,
                           r.route,
                           r.activity_journal_id
                      from route as r),
       fragment2d as (select (st_dump(st_intersection(s.straight, c.hull))).geom as fragment,
                             s.activity_journal_id,
                             c.id as cluster_hull_id,
                             s.route as route
                        from straight as s,
                             cluster_hull as c
                       where st_intersects(s.straight, c.hull)
                         and c.sector_group_id = :sector_group_id),
       startend as (select activity_journal_id,
                           cluster_hull_id,
                           route,
                           fragment,
                           st_linelocatepoint(route, st_startpoint(fragment)) as a,
                           st_linelocatepoint(route, st_endpoint(fragment)) as b
                      from fragment2d),
       fragment3d as (select activity_journal_id,
                             cluster_hull_id,
                             st_linesubstring(route,
                                              least(1, greatest(0, least(a, b))),
                                              least(1, greatest(0, greatest(a, b)))) as fragment
                        from startend),
       lines3d as (select activity_journal_id,
                          cluster_hull_id,
                          fragment
                     from fragment3d
                    where st_geometrytype(fragment) = 'ST_LineString'
                      and st_distance(st_startpoint(fragment), st_endpoint(fragment)) > :min_separation)
insert into cluster_fragment_scratch (cluster_hull_id, activity_journal_id, fragment, length)
select distinct cluster_hull_id, activity_journal_id, fragment, st_length(fragment)
  from lines3d;
    ''')
    log.debug(sql)
    s.connection().execute(sql, sector_group_id=sector_group.id, min_separation=min_separation)
    s.commit()


def delete_tmp_fragments(s, sector_group):
    cluster_hull_ids = s.query(ClusterHull.id). \
        filter(ClusterHull.sector_group_id == sector_group.id)
    s.query(ClusterFragmentScratch). \
        filter(ClusterFragmentScratch.cluster_hull_id.in_(cluster_hull_ids)). \
        delete(synchronize_session=False)


def delete_sectors(s, sector_group):
    from ..pipeline.calculate.cluster import ClusterCalculator
    query = s.query(Sector). \
        filter(Sector.sector_group_id == sector_group.id,
               Sector.owner == short_cls(ClusterCalculator))
    n = query.count()
    if n:
        log.warning(f'Deleting {n} sectors for clusters in {sector_group.title}')
        query.delete(synchronize_session=False)


def identify_sectors(s, sector_group):
    from ..pipeline.calculate.cluster import ClusterCalculator
    sql = text('''
  with median as (select cfs.cluster_hull_id,
                         percentile_disc(0.75) within group (order by cfs.length) as length,
                         count(cfs.id) as n
                    from cluster_fragment_scratch as cfs,
                         cluster_hull as ch
                   where cfs.cluster_hull_id = ch.id
                     and ch.sector_group_id = :sector_group_id
                   group by cluster_hull_id),
       typical as (select f.activity_journal_id as activity_journal_id,
                          m.cluster_hull_id,
                          m.length,
                          f.fragment
                     from median as m,
                          cluster_fragment_scratch as f
                    where m.cluster_hull_id = f.cluster_hull_id
                      and m.length = f.length
                      and n > 2
                      and m.length > 500)
insert into sector (type, sector_group_id, route, distance, owner)
select :sector_type, :sector_group_id, st_force2d(fragment), length / 1000, :owner
  from typical;
''')
    log.debug(sql)
    s.connection().execute(sql, sector_group_id=sector_group.id, sector_type=SectorType.SECTOR,
                           owner=short_cls(ClusterCalculator))
    s.commit()




from abc import abstractmethod
from logging import getLogger

from sqlalchemy import not_
from sqlalchemy.sql.functions import count

from ..pipeline import ProcessPipeline, OwnerInMixin
from ...common.date import time_to_local_timeq, format_dateq
from ...common.log import log_current_exception, log_query
from ...lib import local_time_to_time, to_date
from ...lib.schedule import Schedule
from ...sql import Timestamp, ActivityJournal, ActivityGroup, Interval
from ...sql.types import short_cls
from ...sql.utils import add

log = getLogger(__name__)


class ProcessCalculator(ProcessPipeline): pass


class JournalProcessCalculator(ProcessCalculator):
    '''
    auto-detects missing entries and schedules threads via owner_out.

    provides access to the journal via _get_source.
    '''

    _journal_type = None

    def _missing(self, s):
        source_ids = s.query(Timestamp.source_id).filter(Timestamp.owner == self.owner_out)
        q = self._delimit_missing(
            s.query(self._journal_type.start).
                filter(not_(self._journal_type.id.in_(source_ids))).
                order_by(self._journal_type.start))
        return [time_to_local_timeq(row[0]) for row in q]

    def _delimit_missing(self, q):
        return q

    def _get_source(self, s, time):
        return s.query(self._journal_type).filter(self._journal_type.start == time).one()


class ActivityJournalProcessCalculator(JournalProcessCalculator):

    _journal_type = ActivityJournal


class ActivityGroupProcessCalculator(ActivityJournalProcessCalculator):

    def __init__(self, *args, activity_group=None, **kargs):
        super().__init__(*args, **kargs)
        self.activity_group = self._assert('activity_group', activity_group)

    def _delimit_missing(self, q):
        q = q.join(ActivityGroup).filter(ActivityGroup.name == self.activity_group)
        return log_query(q, 'Missing:')


class DataFrameCalculatorMixin:

    def __init__(self, *args, add_serial=True, **kargs):
        self.__add_serial = add_serial
        super().__init__(*args, **kargs)

    def _run_one(self, missed):
        with self._config.db.session_context() as s:
            log.debug(f'Calculating for {missed}')
            source = self._get_source(s, local_time_to_time(missed))
            with Timestamp(owner=self.owner_out, source=source).on_success(s):
                try:
                    # data may be structured (doesn't have to be simply a dataframe)
                    data = self._read_dataframe(s, source)
                    stats = self._calculate_stats(s, source, data)
                    if stats is not None:
                        loader = self._get_loader(s, add_serial=self.__add_serial)
                        self._copy_results(s, source, loader, stats)
                        loader.load()
                    else:
                        raise Exception('No stats')
                except Exception as e:
                    log.error(f'No statistics on {missed}: {e}')
                    log_current_exception(traceback=True)

    # allow pass-through to JournalProcessCalculator
    # @abstractmethod
    # def _get_source(self, s, time):
    #     raise NotImplementedError()

    @abstractmethod
    def _read_dataframe(self, s, source):
        raise NotImplementedError()

    @abstractmethod
    def _calculate_stats(self, s, source, data):
        raise NotImplementedError()

    @abstractmethod
    def _copy_results(self, s, source, loader, stats):
        raise NotImplementedError()


class IntervalCalculatorMixin:

    def __init__(self, *args, schedule='m', grouped=False, permanent=False, **kargs):
        self.schedule = Schedule(self._assert('schedule', schedule))
        self.grouped = grouped
        self.permanent = permanent
        super().__init__(*args, **kargs)

    def _missing(self, s):
        from .summary import SummaryCalculator
        expected = s.query(ActivityGroup).count() + 1 if self.grouped else 1

        return [format_dateq(start)
                for start in Interval.missing_starts(s, expected, self.schedule, self.owner_out,
                                                     exclude_owners=(SummaryCalculator,))]

    def _run_one(self, missed):
        start = to_date(missed)
        with self._config.db.session_context() as s:
            activity_groups = [None] + (list(s.query(ActivityGroup).all()) if self.grouped else [])
            for activity_group in activity_groups:
                log.debug(f'Activity group: {activity_group}; Schedule: {self.schedule}')
                if s.query(Interval). \
                        filter(Interval.schedule == self.schedule,
                               Interval.owner == self.owner_out,
                               Interval.start == start,
                               Interval.activity_group == activity_group).one_or_none():
                    # we can have some activity groups, but not others
                    log.warning(f'Interval already exists for '
                                f'{activity_group} / {self.schedule} at {missed}')
                else:
                    interval = add(s, Interval(schedule=self.schedule, owner=self.owner_out,
                                               start=start, activity_group=activity_group,
                                               permanent=self.permanent))
                    s.commit()
                    try:
                        data = self._read_data(s, interval)
                        loader = self._get_loader(s, add_serial=False, clear_timestamp=False)
                        self._calculate_results(s, interval, data, loader)
                        loader.load()
                    except Exception as e:
                        log.error(f'No statistics for {missed} due to error ({e})')
                        log_current_exception()

    @abstractmethod
    def _read_data(self, s, interval):
        raise NotImplementedError()

    @abstractmethod
    def _calculate_results(self, s, interval, data, loader):
        raise NotImplementedError()


class RerunWhenNewActivitiesMixin(OwnerInMixin):

    def __init__(self, *args, excess=0, **kwargs):
        super().__init__(*args, **kwargs)
        self.__excess = excess

    def _missing(self, s):
        prev = Timestamp.get(s, self.owner_out)
        if not prev:
            return ['missing']
        prev_ids = s.query(Timestamp.source_id). \
            filter(Timestamp.owner == self.owner_in,
                   Timestamp.time < prev.time)
        after = s.query(count(ActivityJournal.id)). \
            join(ActivityGroup). \
            filter(not_(ActivityJournal.id.in_(prev_ids.cte()))).scalar()
        if self.__excess:
            before = prev_ids.count()
            missing = after > self.__excess * before
            if missing:
                log.info(f'Number of new activities ({after}) exceeds threshold '
                         f'({self.__excess} x {before} = {self.__excess * before})')
            else:
                log.info('No new data')
        else:
            missing = after
            if missing:
                log.info('New data so reprocess')
            else:
                log.info('No new data')
        if missing:
            s.query(Timestamp).filter(Timestamp.owner == self.owner_out).delete(synchronize_session=False)
            return ['missing']
        else:
            return []



from logging import getLogger

from sqlalchemy import create_engine, Integer, Column, union, select
from sqlalchemy.orm import declarative_base, sessionmaker

from ch2 import upload, BASE
from ch2.commands.args import V, bootstrap_db, DEV, UPLOAD, TEXT
from ch2.commands.search import do_search
from ch2.common.args import m, mm
from ch2.config.profiles.acooke import acooke
from tests import LogTestCase, random_test_user, TempDirOnSuccess

log = getLogger(__name__)


class TestSearch(LogTestCase):

    def test_search(self):

        user = random_test_user()
        with TempDirOnSuccess() as f:
            config = bootstrap_db(user, mm(BASE), f, m(V), '5', configurator=acooke)

            with self.assertRaisesRegex(Exception, 'could not be validated'):
                do_search(config.db, ['active-distance > 100'])

            config = bootstrap_db(user, mm(BASE), f, m(V), '5', mm(DEV), UPLOAD,
                                  'data/test/source/personal/2018-03-04-qdp.fit')
            upload(config)
            self.assertTrue(do_search(config.db, ['active-distance > 100']))
            self.assertTrue(do_search(config.db, ['2018-03-04T07:16:33'], cmd=TEXT))
            self.assertTrue(do_search(config.db, ['start > 2018-03-01']))
            self.assertFalse(do_search(config.db, ['start < 2018-03-01']))


class SQLAlchemyTest1(LogTestCase):

    def test_sqlalchemy(self):

        engine = create_engine('sqlite:///:memory:')
        Base = declarative_base()
        Session = sessionmaker(engine)

        class Table(Base):
            __tablename__ = 'table'
            id = Column(Integer, primary_key=True)

        Base.metadata.create_all(engine)

        with Session() as s:
            q1 = select(Table).filter(Table.id == 1)
            q2 = select(Table).filter(Table.id == 2)
            q3 = union(q1, q2).subquery()

        with Session() as s:
            q1 = select(Table).filter(Table.id == 1)
            q2 = select(Table).filter(Table.id == 2)
            cte = union(q1, q2).cte()
            q3 = select(Table).filter(Table.id.in_(cte))
            q4 = select(Table).filter(Table.id.in_(cte))
            q5 = union(q3, q4).subquery()
            log.debug(q5)





import datetime as dt
from collections import namedtuple
from json import loads
from logging import getLogger

import pytz
from math import log10
from sqlalchemy import distinct
from sqlalchemy.sql.functions import count

from .utils import ProcessCalculator
from ..pipeline import LoaderMixin, OwnerInMixin
from ..read.activity import ActivityReader
from ...common.date import round_hour, to_time, now, format_timeq
from ...common.math import is_nan
from ...common.names import TIME_ZERO
from ...data import Statistics, present
from ...data.response import sum_to_hour, calc_response
from ...names import Names as N, SPACE
from ...sql import StatisticJournal, Composite, StatisticName, Source, Constant, CompositeComponent, \
    StatisticJournalType
from ...sql.tables.source import SourceType
from ...sql.utils import add

log = getLogger(__name__)

Response = namedtuple('Response', 'src_owner, title, tau_days, start, scale')

SCALED = 'scaled'


class ResponseCalculator(LoaderMixin, OwnerInMixin, ProcessCalculator):
    '''
    this is hard to do correctly, incrementally.

    for now, we do all or nothing.

    first, we check if the current solution is complete:
    * all sources used (no need to check for gaps - composite chaining should do that)
    * within 3 hours of the current time (if not, we regenerate, padding with zeroes)

    if not complete, we regenerate the whole damn thing.
    '''

    def __init__(self, *args, response_constants=None, prefix=None, **kargs):
        self.response_constant_names = self._assert('response_constants', response_constants)
        self.prefix = self._assert('prefix', prefix)
        super().__init__(*args, **kargs)

    def _startup(self, s):
        self.response_constants = [Constant.from_name(s, name) for name in self.response_constant_names]
        self.responses = [Response(**loads(constant.at(s).value)) for constant in self.response_constants]
        super()._startup(s)
        for constant, response in zip(self.response_constants, self.responses):
            name = self.prefix + SPACE + constant.short_name
            self._provides(s, name, StatisticJournalType.FLOAT, None, None,
                           f'The SHRIMP response for a decay of {response.tau_days} days', title=response.title)

    def _delete(self, s):
        composite_ids = s.query(Composite.id). \
            join(StatisticJournal, Composite.id == StatisticJournal.source_id). \
            join(StatisticName, StatisticJournal.statistic_name_id == StatisticName.id). \
            filter(StatisticName.owner == self.owner_out)
        log.debug(f'Delete query: {composite_ids}')
        n = s.query(count(Source.id)). \
            filter(Source.id.in_(composite_ids)). \
            scalar()
        if n:
            log.warning(f'Deleting {n} Composite sources')
            s.query(Source). \
                filter(Source.id.in_(composite_ids)). \
                delete(synchronize_session=False)
            s.commit()
            Composite.clean(s)

    def _missing(self, s):
        # clean out any gaps by unzipping the chained composites
        Composite.clean(s)
        # range we expect data for
        finish = round_hour(dt.datetime.now(tz=pytz.UTC), up=True)
        missing_sources = self.__missing_sources(s)
        missing_recent = any(self.__missing_recent(s, constant.short_name, finish)
                             for constant in self.response_constants)
        if missing_recent or missing_sources:
            if missing_recent: log.info('Incomplete coverage (so will re-calculate)')
            if missing_sources: log.info('Additional sources (so will re-calculate)')
            self._delete(s)
            start = round_hour(self.__start(s), up=False)
            return [format_timeq(start)]
        else:
            return []

    def __missing_recent(self, s, constant, now):
        log.debug('Searching for missing recent')
        finish = s.query(StatisticJournal.time). \
            join(StatisticName, StatisticJournal.statistic_name_id == StatisticName.id). \
            filter(StatisticName.name == self.prefix + SPACE + constant,
                   StatisticName.owner == self.owner_out). \
            order_by(StatisticJournal.time.desc()).limit(1).one_or_none()
        if finish is None:
            log.debug(f'No data for {constant}')
            return False  # may simply have no data
        if (now - finish[0]).total_seconds() > 3 * 60 * 60:
            log.debug(f'Insufficient data for {constant}')
            return True
        return False

    def __missing_sources(self, s):
        log.debug('Searching for missing sources')
        available = s.query(count(distinct(Source.id))). \
            join(StatisticJournal). \
            join(StatisticName). \
            filter(StatisticName.name == self.prefix + SPACE + N.HR_IMPULSE_10)
        used = s.query(count(distinct(Source.id))). \
            join(CompositeComponent, CompositeComponent.input_source_id == Source.id). \
            join(Composite, Composite.id == CompositeComponent.output_source_id). \
            join(StatisticJournal, StatisticJournal.source_id == Composite.id). \
            join(StatisticName). \
            filter(StatisticName.owner == self.owner_out,
                   Source.type == SourceType.ACTIVITY)
        n_available = available.scalar()
        n_used = used.scalar()
        log.debug(f'Using {n_used} of {n_available} sources')
        return n_used != n_available

    def __start(self, s):
        # avoid constants defined at time 0
        return s.query(StatisticJournal.time). \
            filter(StatisticJournal.time > TIME_ZERO). \
            order_by(StatisticJournal.time.asc()). \
            limit(1).one()[0]  # scalar can return None

    def _run_one(self, missed):
        with self._config.db.session_context() as s:
            data = self.__read_data(s)
            if N.HR_IMPULSE_10 in data.columns and N.COVERAGE in data.columns:
                # coverage is calculated by the loader and seems to reflect records that have location data but not
                # HR data.  so i guess it makes sense to scale.
                # i don't remember why / when i added this - it might have been when using an optical monitor?
                # it seems like a relatively small effect in most cases.
                data.loc[now()] = {N.HR_IMPULSE_10: 0.0, N._src(N.HR_IMPULSE_10): None, N.COVERAGE: 100}
                data[SCALED] = data[N.HR_IMPULSE_10] * 100 / data[N.COVERAGE]
                all_sources = list(self.__make_sources(s, data))
                for constant, response in zip(self.response_constants, self.responses):
                    name = self.prefix + SPACE + constant.short_name
                    log.info(f'Creating values for {response.title} ({name})')
                    imp3600 = sum_to_hour(data, SCALED)
                    params = (log10(response.tau_days * 24),
                              log10(response.start) if response.start > 0 else 1)
                    result = calc_response(imp3600, params) * response.scale
                    loader = self._get_loader(s, add_serial=False)
                    source, sources = None, list(all_sources)
                    for time, value in result.iteritems():
                        # the sources are much more spread out than the response, which is calculated every hour so
                        # that it is smooth.  so we only increment the source when necessary.
                        skipped = 0
                        while sources and time >= sources[0][0]:
                            source = sources.pop(0)[1]
                            skipped += 1
                            if skipped > 1:
                                log.warning(f'Skipping multiple sources at {time}')
                        loader.add_data(name, source, value, time)
                    loader.load()

    def __read_data(self, s):
        from ..owners import ImpulseCalculator
        name = self.prefix + SPACE + N.HR_IMPULSE_10
        df = Statistics(s, with_sources=True).by_name(ImpulseCalculator, name).with_. \
            rename({name: N.HR_IMPULSE_10, N._src(name): N._src(N.HR_IMPULSE_10)}).df
        name = N._cov(N.HEART_RATE)
        df = Statistics(s).by_name(ActivityReader, name).with_. \
            rename({name: N.COVERAGE}).into(df, tolerance='10s')
        if present(df, N.COVERAGE):
            df[N.COVERAGE].fillna(axis='index', method='ffill', inplace=True)
            df[N.COVERAGE].fillna(100, axis='index', inplace=True)
        return df

    def __make_sources(self, s, data):
        # this chains forwards from zero, adding a new composite for each new impulse source.
        # in theory we could reconstruct only missing entries.
        log.info('Creating sources')
        name = N._src(N.HR_IMPULSE_10)
        prev = add(s, Composite(n_components=0))
        yield to_time(0.0), prev
        # find times where the source changes
        changes = data.loc[data[name].ne(data[name].shift())]
        for time, row in changes.iterrows():
            id = row[name]
            if not is_nan(id):
                composite = add(s, Composite(n_components=2))
                add(s, CompositeComponent(input_source_id=id, output_source=composite))
            else:
                composite = add(s, Composite(n_components=1))
            add(s, CompositeComponent(input_source=prev, output_source=composite))
            yield time, composite
            prev = composite
        s.commit()
